@@ -14,9 +14,10 @@ from __future__ import print_function
 import itertools
 import os
 import sqlite3
-import urllib
 from hashlib import md5
 
+from six.moves import urllib
+
 from swift.common.utils import hash_path, storage_directory, \
     Timestamp
 from swift.common.ring import Ring
@@ -100,7 +101,7 @@ def print_ring_locations(ring, datadir, account, container=None, obj=None,
     for node in primary_nodes:
         cmd = 'curl -I -XHEAD "http://%s:%s/%s/%s/%s"' \
             % (node['ip'], node['port'], node['device'], part,
-               urllib.quote(target))
+               urllib.parse.quote(target))
         if policy_index is not None:
             cmd += ' -H "%s: %s"' % ('X-Backend-Storage-Policy-Index',
                                      policy_index)
@@ -108,7 +109,7 @@ def print_ring_locations(ring, datadir, account, container=None, obj=None,
     for node in handoff_nodes:
         cmd = 'curl -I -XHEAD "http://%s:%s/%s/%s/%s"' \
             % (node['ip'], node['port'], node['device'], part,
-               urllib.quote(target))
+               urllib.parse.quote(target))
         if policy_index is not None:
             cmd += ' -H "%s: %s"' % ('X-Backend-Storage-Policy-Index',
                                      policy_index)
@@ -307,7 +308,7 @@ def print_obj_metadata(metadata):
     print_metadata('Other Metadata:', other_metadata)
 
 
-def print_info(db_type, db_file, swift_dir='/etc/swift'):
+def print_info(db_type, db_file, swift_dir='/etc/swift', stale_reads_ok=False):
     if db_type not in ('account', 'container'):
         print("Unrecognized DB type: internal error")
         raise InfoSystemExit()
@@ -317,10 +318,10 @@ def print_info(db_type, db_file, swift_dir='/etc/swift'):
     if not db_file.startswith(('/', './')):
         db_file = './' + db_file  # don't break if the bare db file is given
     if db_type == 'account':
-        broker = AccountBroker(db_file)
+        broker = AccountBroker(db_file, stale_reads_ok=stale_reads_ok)
         datadir = ABDATADIR
     else:
-        broker = ContainerBroker(db_file)
+        broker = ContainerBroker(db_file, stale_reads_ok=stale_reads_ok)
         datadir = CBDATADIR
     try:
         info = broker.get_info()
