@@ -1,4 +1,3 @@
-#! /usr/bin/env python
 # Copyright (c) 2010-2012 OpenStack Foundation
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -18,18 +17,22 @@ from __future__ import print_function
 import logging
 
 from errno import EEXIST
-from itertools import islice, izip
+from itertools import islice
 from operator import itemgetter
 from os import mkdir
 from os.path import basename, abspath, dirname, exists, join as pathjoin
 from sys import argv as sys_argv, exit, stderr, stdout
 from textwrap import wrap
 from time import time
+from datetime import timedelta
 import optparse
 import math
 
+from six.moves import zip as izip
+from six.moves import input
+
 from swift.common import exceptions
-from swift.common.ring import RingBuilder, Ring
+from swift.common.ring import RingBuilder, Ring, RingData
 from swift.common.ring.builder import MAX_BALANCE
 from swift.common.ring.utils import validate_args, \
     validate_and_normalize_ip, build_dev_from_opts, \
@@ -178,7 +181,7 @@ def _parse_add_values(argvish):
     return parsed_devs
 
 
-def _set_weight_values(devs, weight):
+def _set_weight_values(devs, weight, opts):
     if not devs:
         print('Search value matched 0 devices.\n'
               'The on-disk ring builder is unchanged.')
@@ -188,8 +191,9 @@ def _set_weight_values(devs, weight):
         print('Matched more than one device:')
         for dev in devs:
             print('    %s' % format_device(dev))
-        if raw_input('Are you sure you want to update the weight for '
-                     'these %s devices? (y/N) ' % len(devs)) != 'y':
+        if not opts.yes and \
+                input('Are you sure you want to update the weight for '
+                      'these %s devices? (y/N) ' % len(devs)) != 'y':
             print('Aborting device modifications')
             exit(EXIT_ERROR)
 
@@ -219,7 +223,7 @@ def _parse_set_weight_values(argvish):
                 devs.extend(builder.search_devs(
                     parse_search_value(devstr)) or [])
                 weight = float(weightstr)
-                _set_weight_values(devs, weight)
+                _set_weight_values(devs, weight, opts)
         else:
             if len(args) != 1:
                 print(Commands.set_weight.__doc__.strip())
@@ -228,13 +232,13 @@ def _parse_set_weight_values(argvish):
             devs.extend(builder.search_devs(
                 parse_search_values_from_opts(opts)) or [])
             weight = float(args[0])
-            _set_weight_values(devs, weight)
+            _set_weight_values(devs, weight, opts)
     except ValueError as e:
         print(e)
         exit(EXIT_ERROR)
 
 
-def _set_info_values(devs, change):
+def _set_info_values(devs, change, opts):
 
     if not devs:
         print("Search value matched 0 devices.\n"
@@ -245,8 +249,9 @@ def _set_info_values(devs, change):
         print('Matched more than one device:')
         for dev in devs:
             print('    %s' % format_device(dev))
-        if raw_input('Are you sure you want to update the info for '
-                     'these %s devices? (y/N) ' % len(devs)) != 'y':
+        if not opts.yes and \
+                input('Are you sure you want to update the info for '
+                      'these %s devices? (y/N) ' % len(devs)) != 'y':
             print('Aborting device modifications')
             exit(EXIT_ERROR)
 
@@ -290,14 +295,14 @@ def _parse_set_info_values(argvish):
             devs = builder.search_devs(parse_search_value(search_value))
             change = {}
             ip = ''
-            if len(change_value) and change_value[0].isdigit():
+            if change_value and change_value[0].isdigit():
                 i = 1
                 while (i < len(change_value) and
                        change_value[i] in '0123456789.'):
                     i += 1
                 ip = change_value[:i]
                 change_value = change_value[i:]
-            elif len(change_value) and change_value[0] == '[':
+            elif change_value and change_value.startswith('['):
                 i = 1
                 while i < len(change_value) and change_value[i] != ']':
                     i += 1
@@ -315,14 +320,14 @@ def _parse_set_info_values(argvish):
             if change_value.startswith('R'):
                 change_value = change_value[1:]
                 replication_ip = ''
-                if len(change_value) and change_value[0].isdigit():
+                if change_value and change_value[0].isdigit():
                     i = 1
                     while (i < len(change_value) and
                            change_value[i] in '0123456789.'):
                         i += 1
                     replication_ip = change_value[:i]
                     change_value = change_value[i:]
-                elif len(change_value) and change_value[0] == '[':
+                elif change_value and change_value.startswith('['):
                     i = 1
                     while i < len(change_value) and change_value[i] != ']':
                         i += 1
@@ -351,11 +356,11 @@ def _parse_set_info_values(argvish):
             if change_value or not change:
                 raise ValueError('Invalid set info change value: %s' %
                                  repr(argvish[1]))
-            _set_info_values(devs, change)
+            _set_info_values(devs, change, opts)
     else:
         devs = builder.search_devs(parse_search_values_from_opts(opts))
         change = parse_change_values_from_opts(opts)
-        _set_info_values(devs, change)
+        _set_info_values(devs, change, opts)
 
 
 def _parse_remove_values(argvish):
@@ -379,18 +384,19 @@ def _parse_remove_values(argvish):
             devs.extend(builder.search_devs(
                 parse_search_values_from_opts(opts)))
 
-        return devs
+        return (devs, opts)
     except ValueError as e:
         print(e)
         exit(EXIT_ERROR)
 
 
 class Commands(object):
-
+    @staticmethod
     def unknown():
         print('Unknown command: %s' % argv[2])
         exit(EXIT_ERROR)
 
+    @staticmethod
     def create():
         """
 swift-ring-builder <builder_file> create <part_power> <replicas>
@@ -414,10 +420,13 @@ swift-ring-builder <builder_file> create <part_power> <replicas>
         builder.save(builder_file)
         exit(EXIT_SUCCESS)
 
+    @staticmethod
     def default():
         """
 swift-ring-builder <builder_file>
     Shows information about the ring and the devices within.
+    Flags:
+        DEL - marked for removal and will be removed next rebalance.
         """
         print('%s, build version %d' % (builder_file, builder.version))
         regions = 0
@@ -439,34 +448,45 @@ swift-ring-builder <builder_file>
                   builder.parts, builder.replicas, regions, zones, dev_count,
                   balance, dispersion_trailer))
         print('The minimum number of hours before a partition can be '
-              'reassigned is %s' % builder.min_part_hours)
+              'reassigned is %s (%s remaining)' % (
+                  builder.min_part_hours,
+                  timedelta(seconds=builder.min_part_seconds_left)))
         print('The overload factor is %0.2f%% (%.6f)' % (
             builder.overload * 100, builder.overload))
+
+        # compare ring file against builder file
+        if not exists(ring_file):
+            print('Ring file %s not found, '
+                  'probably it hasn\'t been written yet' % ring_file)
+        else:
+            builder_dict = builder.get_ring().to_dict()
+            try:
+                ring_dict = RingData.load(ring_file).to_dict()
+            except Exception as exc:
+                print('Ring file %s is invalid: %r' % (ring_file, exc))
+            else:
+                if builder_dict == ring_dict:
+                    print('Ring file %s is up-to-date' % ring_file)
+                else:
+                    print('Ring file %s is obsolete' % ring_file)
+
         if builder.devs:
+            balance_per_dev = builder._build_balance_per_dev()
             print('Devices:    id  region  zone      ip address  port  '
                   'replication ip  replication port      name '
-                  'weight partitions balance meta')
-            weighted_parts = builder.parts * builder.replicas / \
-                sum(d['weight'] for d in builder.devs if d is not None)
-            for dev in builder.devs:
-                if dev is None:
-                    continue
-                if not dev['weight']:
-                    if dev['parts']:
-                        balance = MAX_BALANCE
-                    else:
-                        balance = 0
-                else:
-                    balance = 100.0 * dev['parts'] / \
-                        (dev['weight'] * weighted_parts) - 100.0
+                  'weight partitions balance flags meta')
+            for dev in builder._iter_devs():
+                flags = 'DEL' if dev in builder._remove_devs else ''
                 print('         %5d %7d %5d %15s %5d %15s %17d %9s %6.02f '
-                      '%10s %7.02f %s' %
+                      '%10s %7.02f %5s %s' %
                       (dev['id'], dev['region'], dev['zone'], dev['ip'],
                        dev['port'], dev['replication_ip'],
                        dev['replication_port'], dev['device'], dev['weight'],
-                       dev['parts'], balance, dev['meta']))
+                       dev['parts'], balance_per_dev[dev['id']], flags,
+                       dev['meta']))
         exit(EXIT_SUCCESS)
 
+    @staticmethod
     def search():
         """
 swift-ring-builder <builder_file> search <search-value>
@@ -517,6 +537,7 @@ swift-ring-builder <builder_file> search
                    dev['meta']))
         exit(EXIT_SUCCESS)
 
+    @staticmethod
     def list_parts():
         """
 swift-ring-builder <builder_file> list_parts <search-value> [<search-value>] ..
@@ -566,6 +587,7 @@ swift-ring-builder <builder_file> list_parts
             print('%9d   %7d' % (partition, count))
         exit(EXIT_SUCCESS)
 
+    @staticmethod
     def add():
         """
 swift-ring-builder <builder_file> add
@@ -616,10 +638,12 @@ swift-ring-builder <builder_file> add
         builder.save(builder_file)
         exit(EXIT_SUCCESS)
 
+    @staticmethod
     def set_weight():
         """
 swift-ring-builder <builder_file> set_weight <search-value> <weight>
     [<search-value> <weight] ...
+    [--yes]
 
 or
 
@@ -627,6 +651,7 @@ swift-ring-builder <builder_file> set_weight
     --region <region> --zone <zone> --ip <ip or hostname> --port <port>
     --replication-ip <r_ip or r_hostname> --replication-port <r_port>
     --device <device_name> --meta <meta> --weight <weight>
+    [--yes]
 
     Where <r_ip>, <r_hostname> and <r_port> are replication ip, hostname
     and port.
@@ -635,6 +660,8 @@ swift-ring-builder <builder_file> set_weight
     Resets the devices' weights. No partitions will be reassigned to or from
     the device until after running 'rebalance'. This is so you can make
     multiple device changes and rebalance them all just once.
+
+    Option --yes assume a yes response to all questions.
         """
         # if len(argv) < 5 or len(argv) % 2 != 1:
         if len(argv) < 5:
@@ -648,11 +675,13 @@ swift-ring-builder <builder_file> set_weight
         builder.save(builder_file)
         exit(EXIT_SUCCESS)
 
+    @staticmethod
     def set_info():
         """
 swift-ring-builder <builder_file> set_info
     <search-value> <ip>:<port>[R<r_ip>:<r_port>]/<device_name>_<meta>
     [<search-value> <ip>:<port>[R<r_ip>:<r_port>]/<device_name>_<meta>] ...
+    [--yes]
 
 or
 
@@ -665,6 +694,7 @@ swift-ring-builder <builder_file> set_info
     --change-replication-port <r_port>
     --change-device <device_name>
     --change-meta <meta>
+    [--yes]
 
     Where <r_ip>, <r_hostname> and <r_port> are replication ip, hostname
     and port.
@@ -677,6 +707,8 @@ swift-ring-builder <builder_file> set_info
     <ip>:<port>/<device_name>_<meta> parameter; just give what you
     want to change. For instance set_info d74 _"snet: 5.6.7.8" would
     just update the meta data for device id 74.
+
+    Option --yes assume a yes response to all questions.
         """
         if len(argv) < 5:
             print(Commands.set_info.__doc__.strip())
@@ -693,16 +725,19 @@ swift-ring-builder <builder_file> set_info
         builder.save(builder_file)
         exit(EXIT_SUCCESS)
 
+    @staticmethod
     def remove():
         """
 swift-ring-builder <builder_file> remove <search-value> [search-value ...]
+    [--yes]
 
 or
 
-swift-ring-builder <builder_file> search
+swift-ring-builder <builder_file> remove
     --region <region> --zone <zone> --ip <ip or hostname> --port <port>
     --replication-ip <r_ip or r_hostname> --replication-port <r_port>
     --device <device_name> --meta <meta> --weight <weight>
+    [--yes]
 
     Where <r_ip>, <r_hostname> and <r_port> are replication ip, hostname
     and port.
@@ -714,6 +749,8 @@ swift-ring-builder <builder_file> search
     remove command. This will not take effect until after running 'rebalance'.
     This is so you can make multiple device changes and rebalance them all just
     once.
+
+    Option --yes assume a yes response to all questions.
         """
         if len(argv) < 4:
             print(Commands.remove.__doc__.strip())
@@ -721,7 +758,7 @@ swift-ring-builder <builder_file> search
             print(parse_search_value.__doc__.strip())
             exit(EXIT_ERROR)
 
-        devs = _parse_remove_values(argv[3:])
+        devs, opts = _parse_remove_values(argv[3:])
 
         if not devs:
             print('Search value matched 0 devices.\n'
@@ -732,8 +769,9 @@ swift-ring-builder <builder_file> search
             print('Matched more than one device:')
             for dev in devs:
                 print('    %s' % format_device(dev))
-            if raw_input('Are you sure you want to remove these %s '
-                         'devices? (y/N) ' % len(devs)) != 'y':
+            if not opts.yes and \
+                    input('Are you sure you want to remove these %s '
+                          'devices? (y/N) ' % len(devs)) != 'y':
                 print('Aborting device removals')
                 exit(EXIT_ERROR)
 
@@ -758,6 +796,7 @@ swift-ring-builder <builder_file> search
         builder.save(builder_file)
         exit(EXIT_SUCCESS)
 
+    @staticmethod
     def rebalance():
         """
 swift-ring-builder <builder_file> rebalance [options]
@@ -791,10 +830,18 @@ swift-ring-builder <builder_file> rebalance [options]
             handler.setFormatter(formatter)
             logger.addHandler(handler)
 
+        if builder.min_part_seconds_left > 0 and not options.force:
+            print('No partitions could be reassigned.')
+            print('The time between rebalances must be at least '
+                  'min_part_hours: %s hours (%s remaining)' % (
+                      builder.min_part_hours,
+                      timedelta(seconds=builder.min_part_seconds_left)))
+            exit(EXIT_WARNING)
+
         devs_changed = builder.devs_changed
         try:
             last_balance = builder.get_balance()
-            parts, balance = builder.rebalance(seed=get_seed(3))
+            parts, balance, removed_devs = builder.rebalance(seed=get_seed(3))
         except exceptions.RingBuilderError as e:
             print('-' * 79)
             print("An error has occurred during ring validation. Common\n"
@@ -804,10 +851,9 @@ swift-ring-builder <builder_file> rebalance [options]
                   (e,))
             print('-' * 79)
             exit(EXIT_ERROR)
-        if not (parts or options.force):
+        if not (parts or options.force or removed_devs):
             print('No partitions could be reassigned.')
-            print('Either none need to be or none can be due to '
-                  'min_part_hours [%s].' % builder.min_part_hours)
+            print('There is no need to do so at this time')
             exit(EXIT_WARNING)
         # If we set device's weight to zero, currently balance will be set
         # special value(MAX_BALANCE) until zero weighted device return all
@@ -863,6 +909,7 @@ swift-ring-builder <builder_file> rebalance [options]
         builder.save(builder_file)
         exit(status)
 
+    @staticmethod
     def dispersion():
         """
 swift-ring-builder <builder_file> dispersion <search_filter> [options]
@@ -872,7 +919,15 @@ swift-ring-builder <builder_file> dispersion <search_filter> [options]
     --verbose option will display dispersion graph broken down by tier
 
     You can filter which tiers are evaluated to drill down using a regex
-    in the optional search_filter arguemnt.
+    in the optional search_filter arguemnt.  i.e.
+
+        swift-ring-builder <builder_file> dispersion "r\d+z\d+$" -v
+
+    ... would only display rows for the zone tiers
+
+        swift-ring-builder <builder_file> dispersion ".*\-[^/]*$" -v
+
+    ... would only display rows for the server tiers
 
     The reports columns are:
 
@@ -910,6 +965,8 @@ swift-ring-builder <builder_file> dispersion <search_filter> [options]
                                    verbose=options.verbose)
         print('Dispersion is %.06f, Balance is %.06f, Overload is %0.2f%%' % (
             builder.dispersion, builder.get_balance(), builder.overload * 100))
+        print('Required overload is %.6f%%' % (
+            builder.get_required_overload() * 100))
         if report['worst_tier']:
             status = EXIT_WARNING
             print('Worst tier is %.06f (%s)' % (report['max_dispersion'],
@@ -947,6 +1004,7 @@ swift-ring-builder <builder_file> dispersion <search_filter> [options]
                 print(template % args)
         exit(status)
 
+    @staticmethod
     def validate():
         """
 swift-ring-builder <builder_file> validate
@@ -955,6 +1013,7 @@ swift-ring-builder <builder_file> validate
         builder.validate()
         exit(EXIT_SUCCESS)
 
+    @staticmethod
     def write_ring():
         """
 swift-ring-builder <builder_file> write_ring
@@ -976,6 +1035,7 @@ swift-ring-builder <builder_file> write_ring
         ring_data.save(ring_file)
         exit(EXIT_SUCCESS)
 
+    @staticmethod
     def write_builder():
         """
 swift-ring-builder <ring_file> write_builder [min_part_hours]
@@ -996,6 +1056,8 @@ swift-ring-builder <ring_file> write_builder [min_part_hours]
             min_part_hours = 24
         ring = Ring(ring_file)
         for dev in ring.devs:
+            if dev is None:
+                continue
             dev.update({
                 'parts': 0,
                 'parts_wanted': 0,
@@ -1018,14 +1080,28 @@ swift-ring-builder <ring_file> write_builder [min_part_hours]
         for parts in builder._replica2part2dev:
             for dev_id in parts:
                 builder.devs[dev_id]['parts'] += 1
-        builder._set_parts_wanted()
         builder.save(builder_file)
 
+    @staticmethod
     def pretend_min_part_hours_passed():
+        """
+swift-ring-builder <builder_file> pretend_min_part_hours_passed
+    Resets the clock on the last time a rebalance happened, thus
+    circumventing the min_part_hours check.
+
+    *****************************
+    USE THIS WITH EXTREME CAUTION
+    *****************************
+
+    If you run this command and deploy rebalanced rings before a replication
+    pass completes, you may introduce unavailability in your cluster. This
+    has an end-user impact.
+        """
         builder.pretend_min_part_hours_passed()
         builder.save(builder_file)
         exit(EXIT_SUCCESS)
 
+    @staticmethod
     def set_min_part_hours():
         """
 swift-ring-builder <builder_file> set_min_part_hours <hours>
@@ -1042,6 +1118,7 @@ swift-ring-builder <builder_file> set_min_part_hours <hours>
         builder.save(builder_file)
         exit(EXIT_SUCCESS)
 
+    @staticmethod
     def set_replicas():
         """
 swift-ring-builder <builder_file> set_replicas <replicas>
@@ -1074,6 +1151,7 @@ swift-ring-builder <builder_file> set_replicas <replicas>
         builder.save(builder_file)
         exit(EXIT_SUCCESS)
 
+    @staticmethod
     def set_overload():
         """
 swift-ring-builder <builder_file> set_overload <overload>[%]
@@ -1120,7 +1198,7 @@ swift-ring-builder <builder_file> set_overload <overload>[%]
 
 def main(arguments=None):
     global argv, backup_dir, builder, builder_file, ring_file
-    if arguments:
+    if arguments is not None:
         argv = arguments
     else:
         argv = sys_argv
@@ -1130,11 +1208,12 @@ def main(arguments=None):
               globals())
         print(Commands.default.__doc__.strip())
         print()
-        cmds = [c for c, f in Commands.__dict__.items()
-                if f.__doc__ and c[0] != '_' and c != 'default']
+        cmds = [c for c in dir(Commands)
+                if getattr(Commands, c).__doc__ and not c.startswith('_') and
+                c != 'default']
         cmds.sort()
         for cmd in cmds:
-            print(Commands.__dict__[cmd].__doc__.strip())
+            print(getattr(Commands, cmd).__doc__.strip())
             print()
         print(parse_search_value.__doc__.strip())
         print()
@@ -1179,13 +1258,9 @@ def main(arguments=None):
     if argv[0].endswith('-safe'):
         try:
             with lock_parent_directory(abspath(builder_file), 15):
-                Commands.__dict__.get(command, Commands.unknown.im_func)()
+                getattr(Commands, command, Commands.unknown)()
         except exceptions.LockTimeout:
             print("Ring/builder dir currently locked.")
             exit(2)
     else:
-        Commands.__dict__.get(command, Commands.unknown.im_func)()
-
-
-if __name__ == '__main__':
-    main()
+        getattr(Commands, command, Commands.unknown)()
