@@ -1,4 +1,4 @@
-# Copyright (c) 2010-2012 OpenStack, LLC.
+# Copyright (c) 2010-2012 OpenStack Foundation
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -15,9 +15,12 @@
 
 import os
 import sys
+import time
 import signal
 from re import sub
 
+import eventlet.debug
+
 from swift.common import utils
 
 
@@ -75,15 +78,33 @@ def run_daemon(klass, conf_file, section_name='', once=False, **kwargs):
                           log_name=kwargs.get('log_name'))
 
     # once on command line (i.e. daemonize=false) will over-ride config
-    once = once or \
-            conf.get('daemonize', 'true').lower() not in utils.TRUE_VALUES
+    once = once or not utils.config_true_value(conf.get('daemonize', 'true'))
 
     # pre-configure logger
     if 'logger' in kwargs:
         logger = kwargs.pop('logger')
     else:
         logger = utils.get_logger(conf, conf.get('log_name', section_name),
-           log_to_console=kwargs.pop('verbose', False), log_route=section_name)
+                                  log_to_console=kwargs.pop('verbose', False),
+                                  log_route=section_name)
+
+    # disable fallocate if desired
+    if utils.config_true_value(conf.get('disable_fallocate', 'no')):
+        utils.disable_fallocate()
+    # set utils.FALLOCATE_RESERVE if desired
+    utils.FALLOCATE_RESERVE, utils.FALLOCATE_IS_PERCENT = \
+        utils.config_fallocate_value(conf.get('fallocate_reserve', '1%'))
+
+    # By default, disable eventlet printing stacktraces
+    eventlet_debug = utils.config_true_value(conf.get('eventlet_debug', 'no'))
+    eventlet.debug.hub_exceptions(eventlet_debug)
+
+    # Ensure TZ environment variable exists to avoid stat('/etc/localtime') on
+    # some platforms. This locks in reported times to the timezone in which
+    # the server first starts running in locations that periodically change
+    # timezones.
+    os.environ['TZ'] = time.strftime("%z", time.gmtime())
+
     try:
         klass(conf).run(once=once, **kwargs)
     except KeyboardInterrupt:
