@@ -16,17 +16,20 @@
 from eventlet import sleep, Timeout
 from eventlet.green import httplib, socket, urllib2
 import json
+import six
+from six.moves import range
+from six.moves import urllib
 import struct
-from sys import exc_info
+from sys import exc_info, exit
 import zlib
 from swift import gettext_ as _
 from time import gmtime, strftime, time
-import urlparse
 from zlib import compressobj
 
-from swift.common.utils import quote
-from swift.common.http import HTTP_NOT_FOUND
+from swift.common.exceptions import ClientException
+from swift.common.http import HTTP_NOT_FOUND, HTTP_MULTIPLE_CHOICES
 from swift.common.swob import Request
+from swift.common.utils import quote
 from swift.common.wsgi import loadapp, pipeline_property
 
 
@@ -150,8 +153,7 @@ class InternalClient(object):
 
     def make_request(
             self, method, path, headers, acceptable_statuses, body_file=None):
-        """
-        Makes a request to Swift with retries.
+        """Makes a request to Swift with retries.
 
         :param method: HTTP method of request.
         :param path: Path of request.
@@ -160,7 +162,7 @@ class InternalClient(object):
         :param body_file: Body file to be passed along with request,
                           defaults to None.
 
-        :returns : Response object on success.
+        :returns: Response object on success.
 
         :raises UnexpectedResponse: Exception raised when make_request() fails
                                     to get a response with an acceptable status
@@ -171,7 +173,7 @@ class InternalClient(object):
         headers = dict(headers)
         headers['user-agent'] = self.user_agent
         resp = exc_type = exc_value = exc_traceback = None
-        for attempt in xrange(self.request_tries):
+        for attempt in range(self.request_tries):
             req = Request.blank(
                 path, environ={'REQUEST_METHOD': method}, headers=headers)
             if body_file is not None:
@@ -193,7 +195,7 @@ class InternalClient(object):
                 _('Unexpected response: %s') % resp.status, resp)
         if exc_type:
             # To make pep8 tool happy, in place of raise t, v, tb:
-            raise exc_type(*exc_value.args), None, exc_traceback
+            six.reraise(exc_type(*exc_value.args), None, exc_traceback)
 
     def _get_metadata(
             self, path, metadata_prefix='', acceptable_statuses=(2,),
@@ -210,8 +212,8 @@ class InternalClient(object):
                                     defaults to (2,).
         :param headers: extra headers to send
 
-        :returns : A dict of metadata with metadata_prefix stripped from keys.
-                   Keys will be lowercase.
+        :returns: A dict of metadata with metadata_prefix stripped from keys.
+                  Keys will be lowercase.
 
         :raises UnexpectedResponse: Exception raised when requests fail
                                     to get a response with an acceptable status
@@ -223,7 +225,7 @@ class InternalClient(object):
         resp = self.make_request('HEAD', path, headers, acceptable_statuses)
         metadata_prefix = metadata_prefix.lower()
         metadata = {}
-        for k, v in resp.headers.iteritems():
+        for k, v in resp.headers.items():
             if k.lower().startswith(metadata_prefix):
                 metadata[k[len(metadata_prefix):].lower()] = v
         return metadata
@@ -254,6 +256,8 @@ class InternalClient(object):
                 (path, quote(marker), quote(end_marker)),
                 {}, acceptable_statuses)
             if not resp.status_int == 200:
+                if resp.status_int >= HTTP_MULTIPLE_CHOICES:
+                    ''.join(resp.app_iter)
                 break
             data = json.loads(resp.body)
             if not data:
@@ -307,7 +311,7 @@ class InternalClient(object):
         """
 
         headers = {}
-        for k, v in metadata.iteritems():
+        for k, v in metadata.items():
             if k.lower().startswith(metadata_prefix):
                 headers[k] = v
             else:
@@ -362,8 +366,7 @@ class InternalClient(object):
 
     def get_account_metadata(
             self, account, metadata_prefix='', acceptable_statuses=(2,)):
-        """
-        Gets account metadata.
+        """Gets account metadata.
 
         :param account: Account on which to get the metadata.
         :param metadata_prefix: Used to filter values from the headers
@@ -372,7 +375,7 @@ class InternalClient(object):
         :param acceptable_statuses: List of status for valid responses,
                                     defaults to (2,).
 
-        :returns : Returns dict of account metadata.  Keys will be lowercase.
+        :returns: Returns dict of account metadata.  Keys will be lowercase.
 
         :raises UnexpectedResponse: Exception raised when requests fail
                                     to get a response with an acceptable status
@@ -413,8 +416,7 @@ class InternalClient(object):
     # container methods
 
     def container_exists(self, account, container):
-        """
-        Checks to see if a container exists.
+        """Checks to see if a container exists.
 
         :param account: The container's account.
         :param container: Container to check.
@@ -424,7 +426,7 @@ class InternalClient(object):
         :raises Exception: Exception is raised when code fails in an
                            unexpected way.
 
-        :returns : True if container exists, false otherwise.
+        :returns: True if container exists, false otherwise.
         """
 
         path = self.make_path(account, container)
@@ -474,8 +476,7 @@ class InternalClient(object):
     def get_container_metadata(
             self, account, container, metadata_prefix='',
             acceptable_statuses=(2,)):
-        """
-        Gets container metadata.
+        """Gets container metadata.
 
         :param account: The container's account.
         :param container: Container to get metadata on.
@@ -485,7 +486,7 @@ class InternalClient(object):
         :param acceptable_statuses: List of status for valid responses,
                                     defaults to (2,).
 
-        :returns : Returns dict of container metadata.  Keys will be lowercase.
+        :returns: Returns dict of container metadata.  Keys will be lowercase.
 
         :raises UnexpectedResponse: Exception raised when requests fail
                                     to get a response with an acceptable status
@@ -575,8 +576,7 @@ class InternalClient(object):
     def get_object_metadata(
             self, account, container, obj, metadata_prefix='',
             acceptable_statuses=(2,), headers=None):
-        """
-        Gets object metadata.
+        """Gets object metadata.
 
         :param account: The object's account.
         :param container: The object's container.
@@ -588,7 +588,7 @@ class InternalClient(object):
                                     defaults to (2,).
         :param headers: extra headers to send with request
 
-        :returns : Dict of object metadata.
+        :returns: Dict of object metadata.
 
         :raises UnexpectedResponse: Exception raised when requests fail
                                     to get a response with an acceptable status
@@ -736,11 +736,25 @@ class SimpleClient(object):
     def base_request(self, method, container=None, name=None, prefix=None,
                      headers=None, proxy=None, contents=None,
                      full_listing=None, logger=None, additional_info=None,
-                     timeout=None):
+                     timeout=None, marker=None):
         # Common request method
         trans_start = time()
         url = self.url
 
+        if full_listing:
+            info, body_data = self.base_request(
+                method, container, name, prefix, headers, proxy,
+                timeout=timeout, marker=marker)
+            listing = body_data
+            while listing:
+                marker = listing[-1]['name']
+                info, listing = self.base_request(
+                    method, container, name, prefix, headers, proxy,
+                    timeout=timeout, marker=marker)
+                if listing:
+                    body_data.extend(listing)
+            return [info, body_data]
+
         if headers is None:
             headers = {}
 
@@ -757,13 +771,17 @@ class SimpleClient(object):
             if prefix:
                 url += '&prefix=%s' % prefix
 
+            if marker:
+                url += '&marker=%s' % quote(marker)
+
         req = urllib2.Request(url, headers=headers, data=contents)
         if proxy:
-            proxy = urlparse.urlparse(proxy)
+            proxy = urllib.parse.urlparse(proxy)
             req.set_proxy(proxy.netloc, proxy.scheme)
         req.get_method = lambda: method
         conn = urllib2.urlopen(req, timeout=timeout)
         body = conn.read()
+        info = conn.info()
         try:
             body_data = json.loads(body)
         except ValueError:
@@ -787,13 +805,13 @@ class SimpleClient(object):
                     url,
                     conn.getcode(),
                     sent_content_length,
-                    conn.info()['content-length'],
+                    info['content-length'],
                     trans_start,
                     trans_stop,
                     trans_stop - trans_start,
                     additional_info
                 )))
-        return [None, body_data]
+        return [info, body_data]
 
     def retry_request(self, method, **kwargs):
         retries = kwargs.pop('retries', self.retries)
@@ -803,9 +821,14 @@ class SimpleClient(object):
             self.attempts += 1
             try:
                 return self.base_request(method, **kwargs)
-            except (socket.error, httplib.HTTPException, urllib2.URLError):
+            except (socket.error, httplib.HTTPException, urllib2.URLError) \
+                    as err:
                 if self.attempts > retries:
-                    raise
+                    if isinstance(err, urllib2.HTTPError):
+                        raise ClientException('Raise too many retries',
+                                              http_status=err.getcode())
+                    else:
+                        raise
             sleep(backoff)
             backoff = min(backoff * 2, self.max_backoff)
 
@@ -827,6 +850,12 @@ class SimpleClient(object):
                                   contents=contents.read(), **kwargs)
 
 
+def head_object(url, **kwargs):
+    """For usage with container sync """
+    client = SimpleClient(url=url)
+    return client.retry_request('HEAD', **kwargs)
+
+
 def put_object(url, **kwargs):
     """For usage with container sync """
     client = SimpleClient(url=url)
