@@ -13,6 +13,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from __future__ import print_function
 import functools
 import errno
 import os
@@ -24,20 +25,25 @@ import re
 from swift import gettext_ as _
 
 from swift.common.utils import search_tree, remove_file, write_file
+from swift.common.exceptions import InvalidPidFileException
 
 SWIFT_DIR = '/etc/swift'
 RUN_DIR = '/var/run/swift'
+PROC_DIR = '/proc'
 
 # auth-server has been removed from ALL_SERVERS, start it explicitly
 ALL_SERVERS = ['account-auditor', 'account-server', 'container-auditor',
                'container-replicator', 'container-reconciler',
                'container-server', 'container-sync',
                'container-updater', 'object-auditor', 'object-server',
-               'object-expirer', 'object-replicator', 'object-updater',
+               'object-expirer', 'object-replicator',
+               'object-reconstructor', 'object-updater',
                'proxy-server', 'account-replicator', 'account-reaper']
 MAIN_SERVERS = ['proxy-server', 'account-server', 'container-server',
                 'object-server']
 REST_SERVERS = [s for s in ALL_SERVERS if s not in MAIN_SERVERS]
+# aliases mapping
+ALIASES = {'all': ALL_SERVERS, 'main': MAIN_SERVERS, 'rest': REST_SERVERS}
 GRACEFUL_SHUTDOWN_SERVERS = MAIN_SERVERS + ['auth-server']
 START_ONCE_SERVERS = REST_SERVERS
 # These are servers that match a type (account-*, container-*, object-*) but
@@ -49,7 +55,7 @@ WARNING_WAIT = 3  # seconds to wait after message that may just be a warning
 
 MAX_DESCRIPTORS = 32768
 MAX_MEMORY = (1024 * 1024 * 1024) * 2  # 2 GB
-MAX_PROCS = 8192  # workers * disks * threads_per_disk, can get high
+MAX_PROCS = 8192  # workers * disks, can get high
 
 
 def setup_env():
@@ -59,22 +65,22 @@ def setup_env():
         resource.setrlimit(resource.RLIMIT_NOFILE,
                            (MAX_DESCRIPTORS, MAX_DESCRIPTORS))
     except ValueError:
-        print _("WARNING: Unable to modify file descriptor limit.  "
-                "Running as non-root?")
+        print(_("WARNING: Unable to modify file descriptor limit.  "
+                "Running as non-root?"))
 
     try:
         resource.setrlimit(resource.RLIMIT_DATA,
                            (MAX_MEMORY, MAX_MEMORY))
     except ValueError:
-        print _("WARNING: Unable to modify memory limit.  "
-                "Running as non-root?")
+        print(_("WARNING: Unable to modify memory limit.  "
+                "Running as non-root?"))
 
     try:
         resource.setrlimit(resource.RLIMIT_NPROC,
                            (MAX_PROCS, MAX_PROCS))
     except ValueError:
-        print _("WARNING: Unable to modify max process limit.  "
-                "Running as non-root?")
+        print(_("WARNING: Unable to modify max process limit.  "
+                "Running as non-root?"))
 
     # Set PYTHON_EGG_CACHE if it isn't already set
     os.environ.setdefault('PYTHON_EGG_CACHE', '/tmp')
@@ -133,6 +139,39 @@ def watch_server_pids(server_pids, interval=1, **kwargs):
             time.sleep(0.1)
 
 
+def safe_kill(pid, sig, name):
+    """Send signal to process and check process name
+
+    : param pid: process id
+    : param sig: signal to send
+    : param name: name to ensure target process
+    """
+
+    # check process name for SIG_DFL
+    if sig == signal.SIG_DFL:
+        try:
+            proc_file = '%s/%d/cmdline' % (PROC_DIR, pid)
+            if os.path.exists(proc_file):
+                with open(proc_file, 'r') as fd:
+                    if name not in fd.read():
+                        # unknown process is using the pid
+                        raise InvalidPidFileException()
+        except IOError:
+            pass
+
+    os.kill(pid, sig)
+
+
+def kill_group(pid, sig):
+    """Send signal to process group
+
+    : param pid: process id
+    : param sig: signal to send
+    """
+    # Negative PID means process group
+    os.kill(-pid, sig)
+
+
 class UnknownCommandError(Exception):
     pass
 
@@ -146,18 +185,17 @@ class Manager(object):
 
     def __init__(self, servers, run_dir=RUN_DIR):
         self.server_names = set()
+        self._default_strict = True
         for server in servers:
-            if server == 'all':
-                self.server_names.update(ALL_SERVERS)
-            elif server == 'main':
-                self.server_names.update(MAIN_SERVERS)
-            elif server == 'rest':
-                self.server_names.update(REST_SERVERS)
+            if server in ALIASES:
+                self.server_names.update(ALIASES[server])
+                self._default_strict = False
             elif '*' in server:
                 # convert glob to regex
                 self.server_names.update([
                     s for s in ALL_SERVERS if
                     re.match(server.replace('*', '.*'), s)])
+                self._default_strict = False
             else:
                 self.server_names.add(server)
 
@@ -184,14 +222,23 @@ class Manager(object):
         setup_env()
         status = 0
 
+        strict = kwargs.get('strict')
+        # if strict not set explicitly
+        if strict is None:
+            strict = self._default_strict
+
         for server in self.servers:
-            server.launch(**kwargs)
+            status += 0 if server.launch(**kwargs) else 1
+
+        if not strict:
+            status = 0
+
         if not kwargs.get('daemon', True):
             for server in self.servers:
                 try:
                     status += server.interact(**kwargs)
                 except KeyboardInterrupt:
-                    print _('\nuser quit')
+                    print(_('\nuser quit'))
                     self.stop(**kwargs)
                     break
         elif kwargs.get('wait', True):
@@ -228,7 +275,7 @@ class Manager(object):
         for server in self.servers:
             signaled_pids = server.stop(**kwargs)
             if not signaled_pids:
-                print _('No %s running') % server
+                print(_('No %s running') % server)
             else:
                 server_pids[server] = signaled_pids
 
@@ -241,18 +288,39 @@ class Manager(object):
         for server, killed_pid in watch_server_pids(server_pids,
                                                     interval=kill_wait,
                                                     **kwargs):
-            print _("%s (%s) appears to have stopped") % (server, killed_pid)
+            print(_("%(server)s (%(pid)s) appears to have stopped") %
+                  {'server': server, 'pid': killed_pid})
             killed_pids.add(killed_pid)
             if not killed_pids.symmetric_difference(signaled_pids):
                 # all processes have been stopped
                 return 0
 
         # reached interval n watch_pids w/o killing all servers
+        kill_after_timeout = kwargs.get('kill_after_timeout', False)
         for server, pids in server_pids.items():
             if not killed_pids.issuperset(pids):
                 # some pids of this server were not killed
-                print _('Waited %s seconds for %s to die; giving up') % (
-                    kill_wait, server)
+                if kill_after_timeout:
+                    print(_('Waited %(kill_wait)s seconds for %(server)s '
+                            'to die; killing') %
+                          {'kill_wait': kill_wait, 'server': server})
+                    # Send SIGKILL to all remaining pids
+                    for pid in set(pids.keys()) - killed_pids:
+                        print(_('Signal %(server)s  pid: %(pid)s  signal: '
+                                '%(signal)s') % {'server': server,
+                                                 'pid': pid,
+                                                 'signal': signal.SIGKILL})
+                        # Send SIGKILL to process group
+                        try:
+                            kill_group(pid, signal.SIGKILL)
+                        except OSError as e:
+                            # PID died before kill_group can take action?
+                            if e.errno != errno.ESRCH:
+                                raise
+                else:
+                    print(_('Waited %(kill_wait)s seconds for %(server)s '
+                            'to die; giving up') %
+                          {'kill_wait': kill_wait, 'server': server})
         return 1
 
     @command
@@ -310,9 +378,8 @@ class Manager(object):
 
         """
         cmd = cmd.lower().replace('-', '_')
-        try:
-            f = getattr(self, cmd)
-        except AttributeError:
+        f = getattr(self, cmd, None)
+        if f is None:
             raise UnknownCommandError(cmd)
         if not hasattr(f, 'publicly_accessible'):
             raise UnknownCommandError(cmd)
@@ -374,6 +441,9 @@ class Server(object):
         except AttributeError:
             return False
 
+    def __ne__(self, other):
+        return not self.__eq__(other)
+
     def get_pid_file_name(self, conf_file):
         """Translate conf_file to a corresponding pid_file
 
@@ -434,13 +504,17 @@ class Server(object):
         if not conf_files:
             # maybe there's a config file(s) out there, but I couldn't find it!
             if not kwargs.get('quiet'):
-                print _('Unable to locate config %sfor %s') % (
-                    ('number %s ' % number if number else ''), self.server)
+                if number:
+                    print(_('Unable to locate config number %(number)s for'
+                            ' %(server)s') %
+                          {'number': number, 'server': self.server})
+                else:
+                    print(_('Unable to locate config for %s') % self.server)
             if kwargs.get('verbose') and not kwargs.get('quiet'):
                 if found_conf_files:
-                    print _('Found configs:')
+                    print(_('Found configs:'))
                 for i, conf_file in enumerate(found_conf_files):
-                    print '  %d) %s' % (i + 1, conf_file)
+                    print('  %d) %s' % (i + 1, conf_file))
 
         return conf_files
 
@@ -468,7 +542,11 @@ class Server(object):
         """Generator, yields (pid_file, pids)
         """
         for pid_file in self.pid_files(**kwargs):
-            yield pid_file, int(open(pid_file).read().strip())
+            try:
+                pid = int(open(pid_file).read().strip())
+            except ValueError:
+                pid = None
+            yield pid_file, pid
 
     def signal_pids(self, sig, **kwargs):
         """Send a signal to pids for this server
@@ -480,19 +558,29 @@ class Server(object):
         """
         pids = {}
         for pid_file, pid in self.iter_pid_files(**kwargs):
+            if not pid:  # Catches None and 0
+                print(_('Removing pid file %s with invalid pid') % pid_file)
+                remove_file(pid_file)
+                continue
             try:
                 if sig != signal.SIG_DFL:
-                    print _('Signal %s  pid: %s  signal: %s') % (self.server,
-                                                                 pid, sig)
-                os.kill(pid, sig)
+                    print(_('Signal %(server)s  pid: %(pid)s  signal: '
+                            '%(signal)s') %
+                          {'server': self.server, 'pid': pid, 'signal': sig})
+                safe_kill(pid, sig, 'swift-%s' % self.server)
+            except InvalidPidFileException as e:
+                if kwargs.get('verbose'):
+                    print(_('Removing pid file %(pid_file)s with wrong pid '
+                            '%(pid)d'), {'pid_file': pid_file, 'pid': pid})
+                remove_file(pid_file)
             except OSError as e:
                 if e.errno == errno.ESRCH:
                     # pid does not exist
                     if kwargs.get('verbose'):
-                        print _("Removing stale pid file %s") % pid_file
+                        print(_("Removing stale pid file %s") % pid_file)
                     remove_file(pid_file)
                 elif e.errno == errno.EPERM:
-                    print _("No permission to signal PID %d") % pid
+                    print(_("No permission to signal PID %d") % pid)
             else:
                 # process exists
                 pids[pid] = pid_file
@@ -537,14 +625,16 @@ class Server(object):
                 kwargs['quiet'] = True
                 conf_files = self.conf_files(**kwargs)
                 if conf_files:
-                    print _("%s #%d not running (%s)") % (self.server, number,
-                                                          conf_files[0])
+                    print(_("%(server)s #%(number)d not running (%(conf)s)") %
+                          {'server': self.server, 'number': number,
+                           'conf': conf_files[0]})
             else:
-                print _("No %s running") % self.server
+                print(_("No %s running") % self.server)
             return 1
         for pid, pid_file in pids.items():
             conf_file = self.get_conf_file_name(pid_file)
-            print _("%s running (%s - %s)") % (self.server, pid, conf_file)
+            print(_("%(server)s running (%(pid)s - %(conf)s)") %
+                  {'server': self.server, 'pid': pid, 'conf': conf_file})
         return 0
 
     def spawn(self, conf_file, once=False, wait=True, daemon=True, **kwargs):
@@ -555,7 +645,7 @@ class Server(object):
         :param wait: boolean, if true capture stdout with a pipe
         :param daemon: boolean, if false ask server to log to console
 
-        :returns : the pid of the spawned process
+        :returns: the pid of the spawned process
         """
         args = [self.cmd, conf_file]
         if once:
@@ -596,7 +686,7 @@ class Server(object):
                 # no-daemon anyway, but this is quieter
                 proc.wait()
             if output:
-                print output
+                print(output)
                 start = time.time()
                 # wait for process to die (output may just be a warning)
                 while time.time() - start < WARNING_WAIT:
@@ -637,13 +727,16 @@ class Server(object):
             # any unstarted instances
             if conf_file in conf_files:
                 already_started = True
-                print _("%s running (%s - %s)") % (self.server, pid, conf_file)
+                print(_("%(server)s running (%(pid)s - %(conf)s)") %
+                      {'server': self.server, 'pid': pid, 'conf': conf_file})
             elif not kwargs.get('number', 0):
                 already_started = True
-                print _("%s running (%s - %s)") % (self.server, pid, pid_file)
+                print(_("%(server)s running (%(pid)s - %(pid_file)s)") %
+                      {'server': self.server, 'pid': pid,
+                       'pid_file': pid_file})
 
         if already_started:
-            print _("%s already started...") % self.server
+            print(_("%s already started...") % self.server)
             return {}
 
         if self.server not in START_ONCE_SERVERS:
@@ -655,13 +748,13 @@ class Server(object):
                 msg = _('Running %s once') % self.server
             else:
                 msg = _('Starting %s') % self.server
-            print '%s...(%s)' % (msg, conf_file)
+            print('%s...(%s)' % (msg, conf_file))
             try:
                 pid = self.spawn(conf_file, **kwargs)
             except OSError as e:
                 if e.errno == errno.ENOENT:
-                    #TODO(clayg): should I check if self.cmd exists earlier?
-                    print _("%s does not exist") % self.cmd
+                    # TODO(clayg): should I check if self.cmd exists earlier?
+                    print(_("%s does not exist") % self.cmd)
                     break
                 else:
                     raise
