@@ -23,6 +23,7 @@ import functools
 import sys
 
 from eventlet import Timeout
+import six
 
 from swift import __canonical_version__ as swift_version
 from swift.common import constraints
@@ -35,7 +36,8 @@ from swift.common.utils import cache_from_env, get_logger, \
 from swift.common.constraints import check_utf8, valid_api_version
 from swift.proxy.controllers import AccountController, ContainerController, \
     ObjectControllerRouter, InfoController
-from swift.proxy.controllers.base import get_container_info, NodeIter
+from swift.proxy.controllers.base import get_container_info, NodeIter, \
+    DEFAULT_RECHECK_CONTAINER_EXISTENCE, DEFAULT_RECHECK_ACCOUNT_EXISTENCE
 from swift.common.swob import HTTPBadRequest, HTTPForbidden, \
     HTTPMethodNotAllowed, HTTPNotFound, HTTPPreconditionFailed, \
     HTTPServerError, HTTPException, Request, HTTPServiceUnavailable
@@ -63,9 +65,13 @@ required_filters = [
                                if pipe.startswith('catch_errors')
                                else [])},
     {'name': 'dlo', 'after_fn': lambda _junk: [
-        'staticweb', 'tempauth', 'keystoneauth',
+        'copy', 'staticweb', 'tempauth', 'keystoneauth',
         'catch_errors', 'gatekeeper', 'proxy_logging']},
     {'name': 'versioned_writes', 'after_fn': lambda _junk: [
+        'slo', 'dlo', 'copy', 'staticweb', 'tempauth',
+        'keystoneauth', 'catch_errors', 'gatekeeper', 'proxy_logging']},
+    # Put copy before dlo, slo and versioned_writes
+    {'name': 'copy', 'after_fn': lambda _junk: [
         'staticweb', 'tempauth', 'keystoneauth',
         'catch_errors', 'gatekeeper', 'proxy_logging']}]
 
@@ -86,8 +92,8 @@ class Application(object):
 
         swift_dir = conf.get('swift_dir', '/etc/swift')
         self.swift_dir = swift_dir
-        self.node_timeout = int(conf.get('node_timeout', 10))
-        self.recoverable_node_timeout = int(
+        self.node_timeout = float(conf.get('node_timeout', 10))
+        self.recoverable_node_timeout = float(
             conf.get('recoverable_node_timeout', self.node_timeout))
         self.conn_timeout = float(conf.get('conn_timeout', 0.5))
         self.client_timeout = int(conf.get('client_timeout', 60))
@@ -101,13 +107,13 @@ class Application(object):
         self.error_suppression_limit = \
             int(conf.get('error_suppression_limit', 10))
         self.recheck_container_existence = \
-            int(conf.get('recheck_container_existence', 60))
+            int(conf.get('recheck_container_existence',
+                         DEFAULT_RECHECK_CONTAINER_EXISTENCE))
         self.recheck_account_existence = \
-            int(conf.get('recheck_account_existence', 60))
+            int(conf.get('recheck_account_existence',
+                         DEFAULT_RECHECK_ACCOUNT_EXISTENCE))
         self.allow_account_management = \
             config_true_value(conf.get('allow_account_management', 'no'))
-        self.object_post_as_copy = \
-            config_true_value(conf.get('object_post_as_copy', 'true'))
         self.container_ring = container_ring or Ring(swift_dir,
                                                      ring_name='container')
         self.account_ring = account_ring or Ring(swift_dir,
@@ -146,8 +152,10 @@ class Application(object):
         self.node_timings = {}
         self.timing_expiry = int(conf.get('timing_expiry', 300))
         self.sorting_method = conf.get('sorting_method', 'shuffle').lower()
-        self.max_large_object_get_time = float(
-            conf.get('max_large_object_get_time', '86400'))
+        self.concurrent_gets = \
+            config_true_value(conf.get('concurrent_gets'))
+        self.concurrency_timeout = float(conf.get('concurrency_timeout',
+                                                  self.conn_timeout))
         value = conf.get('request_node_count', '2 * replicas').lower().split()
         if len(value) == 1:
             rnc_value = int(value[0])
@@ -228,9 +236,10 @@ class Application(object):
         Check the configuration for possible errors
         """
         if self._read_affinity and self.sorting_method != 'affinity':
-            self.logger.warn("sorting_method is set to '%s', not 'affinity'; "
-                             "read_affinity setting will have no effect." %
-                             self.sorting_method)
+            self.logger.warning(
+                "sorting_method is set to '%s', not 'affinity'; "
+                "read_affinity setting will have no effect." %
+                self.sorting_method)
 
     def get_object_ring(self, policy_idx):
         """
@@ -343,7 +352,7 @@ class Application(object):
             try:
                 controller, path_parts = self.get_controller(req)
                 p = req.path_info
-                if isinstance(p, unicode):
+                if isinstance(p, six.text_type):
                     p = p.encode('utf-8')
             except APIVersionError:
                 self.logger.increment('errors')
@@ -373,13 +382,12 @@ class Application(object):
             req.headers['x-trans-id'] = req.environ['swift.trans_id']
             controller.trans_id = req.environ['swift.trans_id']
             self.logger.client_ip = get_remote_client(req)
-            try:
-                handler = getattr(controller, req.method)
-                getattr(handler, 'publicly_accessible')
-            except AttributeError:
-                allowed_methods = getattr(controller, 'allowed_methods', set())
-                return HTTPMethodNotAllowed(
-                    request=req, headers={'Allow': ', '.join(allowed_methods)})
+
+            if req.method not in controller.allowed_methods:
+                return HTTPMethodNotAllowed(request=req, headers={
+                    'Allow': ', '.join(controller.allowed_methods)})
+            handler = getattr(controller, req.method)
+
             old_authorize = None
             if 'swift.authorize' in req.environ:
                 # We call authorize before the handler, always. If authorized,
@@ -388,8 +396,7 @@ class Application(object):
                 # controller's method indicates it'd like to gather more
                 # information and try again later.
                 resp = req.environ['swift.authorize'](req)
-                if not resp and not req.headers.get('X-Copy-From-Account') \
-                        and not req.headers.get('Destination-Account'):
+                if not resp:
                     # No resp means authorized, no delayed recheck required.
                     old_authorize = req.environ['swift.authorize']
                 else:
@@ -400,7 +407,7 @@ class Application(object):
             # Save off original request method (GET, POST, etc.) in case it
             # gets mutated during handling.  This way logging can display the
             # method the client actually sent.
-            req.environ['swift.orig_req_method'] = req.method
+            req.environ.setdefault('swift.orig_req_method', req.method)
             try:
                 if old_authorize:
                     req.environ.pop('swift.authorize', None)
@@ -502,7 +509,7 @@ class Application(object):
         """
         self._incr_node_errors(node)
         self.logger.error(_('%(msg)s %(ip)s:%(port)s/%(device)s'),
-                          {'msg': msg, 'ip': node['ip'],
+                          {'msg': msg.decode('utf-8'), 'ip': node['ip'],
                           'port': node['port'], 'device': node['device']})
 
     def iter_nodes(self, ring, partition, node_iter=None):
@@ -528,7 +535,7 @@ class Application(object):
               ' re: %(info)s'),
             {'type': typ, 'ip': node['ip'],
              'port': node['port'], 'device': node['device'],
-             'info': additional_info},
+             'info': additional_info.decode('utf-8')},
             **kwargs)
 
     def modify_wsgi_pipeline(self, pipe):
