@@ -15,12 +15,13 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import unittest
+import unittest2
 import json
 from uuid import uuid4
-from nose import SkipTest
-from string import letters
+from unittest2 import SkipTest
+from string import ascii_letters
 
+from six.moves import range
 from swift.common.middleware.acl import format_acl
 
 from test.functional import check_response, retry, requires_acls, \
@@ -28,7 +29,15 @@ from test.functional import check_response, retry, requires_acls, \
 import test.functional as tf
 
 
-class TestAccount(unittest.TestCase):
+def setUpModule():
+    tf.setup_package()
+
+
+def tearDownModule():
+    tf.teardown_package()
+
+
+class TestAccount(unittest2.TestCase):
 
     def setUp(self):
         self.max_meta_count = load_constraint('max_meta_count')
@@ -88,22 +97,22 @@ class TestAccount(unittest.TestCase):
         self.assertEqual(resp.status, 204)
         resp = retry(head)
         resp.read()
-        self.assert_(resp.status in (200, 204), resp.status)
+        self.assertIn(resp.status, (200, 204))
         self.assertEqual(resp.getheader('x-account-meta-test'), None)
         resp = retry(get)
         resp.read()
-        self.assert_(resp.status in (200, 204), resp.status)
+        self.assertIn(resp.status, (200, 204))
         self.assertEqual(resp.getheader('x-account-meta-test'), None)
         resp = retry(post, 'Value')
         resp.read()
         self.assertEqual(resp.status, 204)
         resp = retry(head)
         resp.read()
-        self.assert_(resp.status in (200, 204), resp.status)
+        self.assertIn(resp.status, (200, 204))
         self.assertEqual(resp.getheader('x-account-meta-test'), 'Value')
         resp = retry(get)
         resp.read()
-        self.assert_(resp.status in (200, 204), resp.status)
+        self.assertIn(resp.status, (200, 204))
         self.assertEqual(resp.getheader('x-account-meta-test'), 'Value')
 
     def test_invalid_acls(self):
@@ -118,7 +127,7 @@ class TestAccount(unittest.TestCase):
         # needs to be an acceptable header size
         num_keys = 8
         max_key_size = load_constraint('max_header_size') / num_keys
-        acl = {'admin': [c * max_key_size for c in letters[:num_keys]]}
+        acl = {'admin': [c * max_key_size for c in ascii_letters[:num_keys]]}
         headers = {'x-account-access-control': format_acl(
             version=2, acl_dict=acl)}
         resp = retry(post, headers=headers, use_account=1)
@@ -126,7 +135,8 @@ class TestAccount(unittest.TestCase):
         self.assertEqual(resp.status, 400)
 
         # and again a touch smaller
-        acl = {'admin': [c * max_key_size for c in letters[:num_keys - 1]]}
+        acl = {'admin': [c * max_key_size for c
+                         in ascii_letters[:num_keys - 1]]}
         headers = {'x-account-access-control': format_acl(
             version=2, acl_dict=acl)}
         resp = retry(post, headers=headers, use_account=1)
@@ -189,7 +199,7 @@ class TestAccount(unittest.TestCase):
         # cannot read account
         resp = retry(get, use_account=3)
         resp.read()
-        self.assertEquals(resp.status, 403)
+        self.assertEqual(resp.status, 403)
 
         # grant read access
         acl_user = tf.swift_test_user[2]
@@ -203,7 +213,7 @@ class TestAccount(unittest.TestCase):
         # read-only can read account headers
         resp = retry(get, use_account=3)
         resp.read()
-        self.assert_(resp.status in (200, 204))
+        self.assertIn(resp.status, (200, 204))
         # but not acls
         self.assertEqual(resp.getheader('X-Account-Access-Control'), None)
 
@@ -220,7 +230,7 @@ class TestAccount(unittest.TestCase):
         self.assertEqual(resp.status, 204)
         resp = retry(get, use_account=3)
         resp.read()
-        self.assert_(resp.status in (200, 204))
+        self.assertIn(resp.status, (200, 204))
         self.assertEqual(resp.getheader('X-Account-Meta-Test'), 'value')
 
     @requires_acls
@@ -240,7 +250,7 @@ class TestAccount(unittest.TestCase):
         # cannot read account
         resp = retry(get, use_account=3)
         resp.read()
-        self.assertEquals(resp.status, 403)
+        self.assertEqual(resp.status, 403)
 
         # grant read-write access
         acl_user = tf.swift_test_user[2]
@@ -254,7 +264,7 @@ class TestAccount(unittest.TestCase):
         # read-write can read account headers
         resp = retry(get, use_account=3)
         resp.read()
-        self.assert_(resp.status in (200, 204))
+        self.assertIn(resp.status, (200, 204))
         # but not acls
         self.assertEqual(resp.getheader('X-Account-Access-Control'), None)
 
@@ -281,7 +291,7 @@ class TestAccount(unittest.TestCase):
         # cannot read account
         resp = retry(get, use_account=3)
         resp.read()
-        self.assertEquals(resp.status, 403)
+        self.assertEqual(resp.status, 403)
 
         # grant admin access
         acl_user = tf.swift_test_user[2]
@@ -295,7 +305,7 @@ class TestAccount(unittest.TestCase):
         # admin can read account headers
         resp = retry(get, use_account=3)
         resp.read()
-        self.assert_(resp.status in (200, 204))
+        self.assertIn(resp.status, (200, 204))
         # including acls
         self.assertEqual(resp.getheader('X-Account-Access-Control'),
                          acl_json_str)
@@ -308,7 +318,7 @@ class TestAccount(unittest.TestCase):
         self.assertEqual(resp.status, 204)
         resp = retry(get, use_account=3)
         resp.read()
-        self.assert_(resp.status in (200, 204))
+        self.assertIn(resp.status, (200, 204))
         self.assertEqual(resp.getheader('X-Account-Meta-Test'), value)
 
         # admin can even revoke their own access
@@ -320,7 +330,7 @@ class TestAccount(unittest.TestCase):
         # and again, cannot read account
         resp = retry(get, use_account=3)
         resp.read()
-        self.assertEquals(resp.status, 403)
+        self.assertEqual(resp.status, 403)
 
     @requires_acls
     def test_protected_tempurl(self):
@@ -358,8 +368,9 @@ class TestAccount(unittest.TestCase):
         # read-only tester3 can read account metadata
         resp = retry(get, use_account=3)
         resp.read()
-        self.assert_(resp.status in (200, 204),
-                     'Expected status in (200, 204), got %s' % resp.status)
+        self.assertTrue(
+            resp.status in (200, 204),
+            'Expected status in (200, 204), got %s' % resp.status)
         self.assertEqual(resp.getheader('X-Account-Meta-Test'), value)
         # but not temp-url-key
         self.assertEqual(resp.getheader('X-Account-Meta-Temp-Url-Key'), None)
@@ -376,8 +387,9 @@ class TestAccount(unittest.TestCase):
         # read-write tester3 can read account metadata
         resp = retry(get, use_account=3)
         resp.read()
-        self.assert_(resp.status in (200, 204),
-                     'Expected status in (200, 204), got %s' % resp.status)
+        self.assertTrue(
+            resp.status in (200, 204),
+            'Expected status in (200, 204), got %s' % resp.status)
         self.assertEqual(resp.getheader('X-Account-Meta-Test'), value)
         # but not temp-url-key
         self.assertEqual(resp.getheader('X-Account-Meta-Temp-Url-Key'), None)
@@ -394,8 +406,9 @@ class TestAccount(unittest.TestCase):
         # admin tester3 can read account metadata
         resp = retry(get, use_account=3)
         resp.read()
-        self.assert_(resp.status in (200, 204),
-                     'Expected status in (200, 204), got %s' % resp.status)
+        self.assertTrue(
+            resp.status in (200, 204),
+            'Expected status in (200, 204), got %s' % resp.status)
         self.assertEqual(resp.getheader('X-Account-Meta-Test'), value)
         # including temp-url-key
         self.assertEqual(resp.getheader('X-Account-Meta-Temp-Url-Key'),
@@ -411,8 +424,9 @@ class TestAccount(unittest.TestCase):
         self.assertEqual(resp.status, 204)
         resp = retry(get, use_account=3)
         resp.read()
-        self.assert_(resp.status in (200, 204),
-                     'Expected status in (200, 204), got %s' % resp.status)
+        self.assertTrue(
+            resp.status in (200, 204),
+            'Expected status in (200, 204), got %s' % resp.status)
         self.assertEqual(resp.getheader('X-Account-Meta-Temp-Url-Key'),
                          secret)
 
@@ -688,17 +702,17 @@ class TestAccount(unittest.TestCase):
         if (tf.web_front_end == 'integral'):
             resp = retry(post, uni_key, '1')
             resp.read()
-            self.assertTrue(resp.status in (201, 204))
+            self.assertIn(resp.status, (201, 204))
             resp = retry(head)
             resp.read()
-            self.assert_(resp.status in (200, 204), resp.status)
+            self.assertIn(resp.status, (200, 204))
             self.assertEqual(resp.getheader(uni_key.encode('utf-8')), '1')
         resp = retry(post, 'X-Account-Meta-uni', uni_value)
         resp.read()
         self.assertEqual(resp.status, 204)
         resp = retry(head)
         resp.read()
-        self.assert_(resp.status in (200, 204), resp.status)
+        self.assertIn(resp.status, (200, 204))
         self.assertEqual(resp.getheader('X-Account-Meta-uni'),
                          uni_value.encode('utf-8'))
         if (tf.web_front_end == 'integral'):
@@ -707,7 +721,7 @@ class TestAccount(unittest.TestCase):
             self.assertEqual(resp.status, 204)
             resp = retry(head)
             resp.read()
-            self.assert_(resp.status in (200, 204), resp.status)
+            self.assertIn(resp.status, (200, 204))
             self.assertEqual(resp.getheader(uni_key.encode('utf-8')),
                              uni_value.encode('utf-8'))
 
@@ -729,14 +743,14 @@ class TestAccount(unittest.TestCase):
         self.assertEqual(resp.status, 204)
         resp = retry(head)
         resp.read()
-        self.assert_(resp.status in (200, 204), resp.status)
+        self.assertIn(resp.status, (200, 204))
         self.assertEqual(resp.getheader('x-account-meta-one'), '1')
         resp = retry(post, 'X-Account-Meta-Two', '2')
         resp.read()
         self.assertEqual(resp.status, 204)
         resp = retry(head)
         resp.read()
-        self.assert_(resp.status in (200, 204), resp.status)
+        self.assertIn(resp.status, (200, 204))
         self.assertEqual(resp.getheader('x-account-meta-one'), '1')
         self.assertEqual(resp.getheader('x-account-meta-two'), '2')
 
@@ -790,13 +804,13 @@ class TestAccount(unittest.TestCase):
         resp = retry(post, headers)
 
         headers = {}
-        for x in xrange(self.max_meta_count):
+        for x in range(self.max_meta_count):
             headers['X-Account-Meta-%d' % x] = 'v'
         resp = retry(post, headers)
         resp.read()
         self.assertEqual(resp.status, 204)
         headers = {}
-        for x in xrange(self.max_meta_count + 1):
+        for x in range(self.max_meta_count + 1):
             headers['X-Account-Meta-%d' % x] = 'v'
         resp = retry(post, headers)
         resp.read()
@@ -827,14 +841,29 @@ class TestAccount(unittest.TestCase):
         resp = retry(post, headers)
         resp.read()
         self.assertEqual(resp.status, 204)
+        # this POST includes metadata size that is over limit
         headers['X-Account-Meta-k'] = \
-            'v' * (self.max_meta_overall_size - size)
+            'x' * (self.max_meta_overall_size - size)
+        resp = retry(post, headers)
+        resp.read()
+        self.assertEqual(resp.status, 400)
+        # this POST would be ok and the aggregate backend metadata
+        # size is on the border
+        headers = {'X-Account-Meta-k':
+                   'y' * (self.max_meta_overall_size - size - 1)}
+        resp = retry(post, headers)
+        resp.read()
+        self.assertEqual(resp.status, 204)
+        # this last POST would be ok by itself but takes the aggregate
+        # backend metadata size over limit
+        headers = {'X-Account-Meta-k':
+                   'z' * (self.max_meta_overall_size - size)}
         resp = retry(post, headers)
         resp.read()
         self.assertEqual(resp.status, 400)
 
 
-class TestAccountInNonDefaultDomain(unittest.TestCase):
+class TestAccountInNonDefaultDomain(unittest2.TestCase):
     def setUp(self):
         if tf.skip or tf.skip2 or tf.skip_if_not_v3:
             raise SkipTest('AUTH VERSION 3 SPECIFIC TEST')
@@ -859,8 +888,8 @@ class TestAccountInNonDefaultDomain(unittest.TestCase):
         resp = retry(head, use_account=4)
         resp.read()
         self.assertEqual(resp.status, 204)
-        self.assertTrue('X-Account-Project-Domain-Id' in resp.headers)
+        self.assertIn('X-Account-Project-Domain-Id', resp.headers)
 
 
 if __name__ == '__main__':
-    unittest.main()
+    unittest2.main()
