@@ -15,53 +15,103 @@
 
 """ Swift tests """
 
+from __future__ import print_function
 import os
 import copy
 import logging
 import errno
+from six.moves import range
 import sys
 from contextlib import contextmanager, closing
 from collections import defaultdict, Iterable
+import itertools
 from numbers import Number
 from tempfile import NamedTemporaryFile
 import time
+import eventlet
 from eventlet.green import socket
 from tempfile import mkdtemp
 from shutil import rmtree
+import signal
+import json
+
+
+from swift.common.utils import Timestamp, NOTICE
 from test import get_config
-from swift.common.utils import config_true_value, LogAdapter
+from swift.common import utils
+from swift.common.header_key_dict import HeaderKeyDict
 from swift.common.ring import Ring, RingData
 from hashlib import md5
-from eventlet import sleep, Timeout
 import logging.handlers
-from httplib import HTTPException
+
+from six.moves.http_client import HTTPException
 from swift.common import storage_policy
+from swift.common.storage_policy import (StoragePolicy, ECStoragePolicy,
+                                         VALID_EC_TYPES)
 import functools
-import cPickle as pickle
+import six.moves.cPickle as pickle
 from gzip import GzipFile
 import mock as mocklib
+import inspect
+
+EMPTY_ETAG = md5().hexdigest()
+
+# try not to import this module from swift
+if not os.path.basename(sys.argv[0]).startswith('swift'):
+    # never patch HASH_PATH_SUFFIX AGAIN!
+    utils.HASH_PATH_SUFFIX = 'endcap'
 
 
-def patch_policies(thing_or_policies=None, legacy_only=False):
-    if legacy_only:
-        default_policies = [storage_policy.StoragePolicy(
-            0, 'legacy', True, object_ring=FakeRing())]
-    else:
-        default_policies = [
-            storage_policy.StoragePolicy(
-                0, 'nulo', True, object_ring=FakeRing()),
-            storage_policy.StoragePolicy(
-                1, 'unu', object_ring=FakeRing()),
-        ]
+EC_TYPE_PREFERENCE = [
+    'liberasurecode_rs_vand',
+    'jerasure_rs_vand',
+]
+for eclib_name in EC_TYPE_PREFERENCE:
+    if eclib_name in VALID_EC_TYPES:
+        break
+else:
+    raise SystemExit('ERROR: unable to find suitable PyECLib type'
+                     ' (none of %r found in %r)' % (
+                         EC_TYPE_PREFERENCE,
+                         VALID_EC_TYPES,
+                     ))
+DEFAULT_TEST_EC_TYPE = eclib_name
 
-    thing_or_policies = thing_or_policies or default_policies
 
+def patch_policies(thing_or_policies=None, legacy_only=False,
+                   with_ec_default=False, fake_ring_args=None):
     if isinstance(thing_or_policies, (
             Iterable, storage_policy.StoragePolicyCollection)):
-        return PatchPolicies(thing_or_policies)
+        return PatchPolicies(thing_or_policies, fake_ring_args=fake_ring_args)
+
+    if legacy_only:
+        default_policies = [
+            StoragePolicy(0, name='legacy', is_default=True),
+        ]
+        default_ring_args = [{}]
+    elif with_ec_default:
+        default_policies = [
+            ECStoragePolicy(0, name='ec', is_default=True,
+                            ec_type=DEFAULT_TEST_EC_TYPE, ec_ndata=10,
+                            ec_nparity=4, ec_segment_size=4096),
+            StoragePolicy(1, name='unu'),
+        ]
+        default_ring_args = [{'replicas': 14}, {}]
     else:
-        # it's a thing!
-        return PatchPolicies(default_policies)(thing_or_policies)
+        default_policies = [
+            StoragePolicy(0, name='nulo', is_default=True),
+            StoragePolicy(1, name='unu'),
+        ]
+        default_ring_args = [{}, {}]
+
+    fake_ring_args = fake_ring_args or default_ring_args
+    decorator = PatchPolicies(default_policies, fake_ring_args=fake_ring_args)
+
+    if not thing_or_policies:
+        return decorator
+    else:
+        # it's a thing, we return the wrapped thing instead of the decorator
+        return decorator(thing_or_policies)
 
 
 class PatchPolicies(object):
@@ -71,11 +121,33 @@ class PatchPolicies(object):
     patched yet)
     """
 
-    def __init__(self, policies):
+    def __init__(self, policies, fake_ring_args=None):
         if isinstance(policies, storage_policy.StoragePolicyCollection):
             self.policies = policies
         else:
             self.policies = storage_policy.StoragePolicyCollection(policies)
+        self.fake_ring_args = fake_ring_args or [None] * len(self.policies)
+
+    def _setup_rings(self):
+        """
+        Our tests tend to use the policies rings like their own personal
+        playground - which can be a problem in the particular case of a
+        patched TestCase class where the FakeRing objects are scoped in the
+        call to the patch_policies wrapper outside of the TestCase instance
+        which can lead to some bled state.
+
+        To help tests get better isolation without having to think about it,
+        here we're capturing the args required to *build* a new FakeRing
+        instances so we can ensure each test method gets a clean ring setup.
+
+        The TestCase can always "tweak" these fresh rings in setUp - or if
+        they'd prefer to get the same "reset" behavior with custom FakeRing's
+        they can pass in their own fake_ring_args to patch_policies instead of
+        setting the object_ring on the policy definitions.
+        """
+        for policy, fake_ring_arg in zip(self.policies, self.fake_ring_args):
+            if fake_ring_arg is not None:
+                policy.object_ring = FakeRing(**fake_ring_arg)
 
     def __call__(self, thing):
         if isinstance(thing, type):
@@ -84,24 +156,33 @@ class PatchPolicies(object):
             return self._patch_method(thing)
 
     def _patch_class(self, cls):
+        """
+        Creating a new class that inherits from decorated class is the more
+        common way I've seen class decorators done - but it seems to cause
+        infinite recursion when super is called from inside methods in the
+        decorated class.
+        """
 
-        class NewClass(cls):
+        orig_setUp = cls.setUp
+        orig_tearDown = cls.tearDown
 
-            already_patched = False
+        def setUp(cls_self):
+            self._orig_POLICIES = storage_policy._POLICIES
+            if not getattr(cls_self, '_policies_patched', False):
+                storage_policy._POLICIES = self.policies
+                self._setup_rings()
+                cls_self._policies_patched = True
 
-            def setUp(cls_self):
-                self._orig_POLICIES = storage_policy._POLICIES
-                if not cls_self.already_patched:
-                    storage_policy._POLICIES = self.policies
-                    cls_self.already_patched = True
-                super(NewClass, cls_self).setUp()
+            orig_setUp(cls_self)
 
-            def tearDown(cls_self):
-                super(NewClass, cls_self).tearDown()
-                storage_policy._POLICIES = self._orig_POLICIES
+        def tearDown(cls_self):
+            orig_tearDown(cls_self)
+            storage_policy._POLICIES = self._orig_POLICIES
 
-        NewClass.__name__ = cls.__name__
-        return NewClass
+        cls.setUp = setUp
+        cls.tearDown = tearDown
+
+        return cls
 
     def _patch_method(self, f):
         @functools.wraps(f)
@@ -109,6 +190,7 @@ class PatchPolicies(object):
             self._orig_POLICIES = storage_policy._POLICIES
             try:
                 storage_policy._POLICIES = self.policies
+                self._setup_rings()
                 return f(*args, **kwargs)
             finally:
                 storage_policy._POLICIES = self._orig_POLICIES
@@ -124,36 +206,27 @@ class PatchPolicies(object):
 
 class FakeRing(Ring):
 
-    def __init__(self, replicas=3, max_more_nodes=0, part_power=0):
-        """
-        :param part_power: make part calculation based on the path
-
-        If you set a part_power when you setup your FakeRing the parts you get
-        out of ring methods will actually be based on the path - otherwise we
-        exercise the real ring code, but ignore the result and return 1.
-        """
+    def __init__(self, replicas=3, max_more_nodes=0, part_power=0,
+                 base_port=1000):
+        self._base_port = base_port
+        self.max_more_nodes = max_more_nodes
+        self._part_shift = 32 - part_power
         # 9 total nodes (6 more past the initial 3) is the cap, no matter if
         # this is set higher, or R^2 for R replicas
         self.set_replicas(replicas)
-        self.max_more_nodes = max_more_nodes
-        self._part_shift = 32 - part_power
         self._reload()
 
     def _reload(self):
         self._rtime = time.time()
 
-    def clear_errors(self):
-        for dev in self.devs:
-            for key in ('errors', 'last_error'):
-                dev.pop(key, None)
-
     def set_replicas(self, replicas):
         self.replicas = replicas
         self._devs = []
         for x in range(self.replicas):
             ip = '10.0.0.%s' % x
-            port = 1000 + x
-            self._devs.append({
+            port = self._base_port + x
+            # round trip through json to ensure unicode like real rings
+            self._devs.append(json.loads(json.dumps({
                 'ip': ip,
                 'replication_ip': ip,
                 'port': port,
@@ -162,21 +235,21 @@ class FakeRing(Ring):
                 'zone': x % 3,
                 'region': x % 2,
                 'id': x,
-            })
+            })))
 
     @property
     def replica_count(self):
         return self.replicas
 
     def _get_part_nodes(self, part):
-        return list(self._devs)
+        return [dict(node, index=i) for i, node in enumerate(list(self._devs))]
 
     def get_more_nodes(self, part):
-        # replicas^2 is the true cap
-        for x in xrange(self.replicas, min(self.replicas + self.max_more_nodes,
-                                           self.replicas * self.replicas)):
+        for x in range(self.replicas, (self.replicas + self.max_more_nodes)):
             yield {'ip': '10.0.0.%s' % x,
-                   'port': 1000 + x,
+                   'replication_ip': '10.0.0.%s' % x,
+                   'port': self._base_port + x,
+                   'replication_port': self._base_port + x,
                    'device': 'sda',
                    'zone': x % 3,
                    'region': x % 2,
@@ -188,9 +261,9 @@ def write_fake_ring(path, *devs):
     Pretty much just a two node, two replica, 2 part power ring...
     """
     dev1 = {'id': 0, 'zone': 0, 'device': 'sda1', 'ip': '127.0.0.1',
-            'port': 6000}
+            'port': 6200}
     dev2 = {'id': 0, 'zone': 0, 'device': 'sdb1', 'ip': '127.0.0.1',
-            'port': 6000}
+            'port': 6200}
 
     dev1_updates, dev2_updates = devs or ({}, {})
 
@@ -204,6 +277,48 @@ def write_fake_ring(path, *devs):
         pickle.dump(RingData(replica2part2dev_id, devs, part_shift), f)
 
 
+class FabricatedRing(Ring):
+    """
+    When a FakeRing just won't do - you can fabricate one to meet
+    your tests needs.
+    """
+
+    def __init__(self, replicas=6, devices=8, nodes=4, port=6200,
+                 part_power=4):
+        self.devices = devices
+        self.nodes = nodes
+        self.port = port
+        self.replicas = 6
+        self.part_power = part_power
+        self._part_shift = 32 - self.part_power
+        self._reload()
+
+    def _reload(self, *args, **kwargs):
+        self._rtime = time.time() * 2
+        if hasattr(self, '_replica2part2dev_id'):
+            return
+        self._devs = [{
+            'region': 1,
+            'zone': 1,
+            'weight': 1.0,
+            'id': i,
+            'device': 'sda%d' % i,
+            'ip': '10.0.0.%d' % (i % self.nodes),
+            'replication_ip': '10.0.0.%d' % (i % self.nodes),
+            'port': self.port,
+            'replication_port': self.port,
+        } for i in range(self.devices)]
+
+        self._replica2part2dev_id = [
+            [None] * 2 ** self.part_power
+            for i in range(self.replicas)
+        ]
+        dev_ids = itertools.cycle(range(self.devices))
+        for p in range(2 ** self.part_power):
+            for r in range(self.replicas):
+                self._replica2part2dev_id[r][p] = next(dev_ids)
+
+
 class FakeMemcache(object):
 
     def __init__(self):
@@ -361,8 +476,14 @@ class UnmockTimeModule(object):
 logging.time = UnmockTimeModule()
 
 
-class FakeLogger(logging.Logger):
-    # a thread safe logger
+class WARN_DEPRECATED(Exception):
+    def __init__(self, msg):
+        self.msg = msg
+        print(self.msg)
+
+
+class FakeLogger(logging.Logger, object):
+    # a thread safe fake logger
 
     def __init__(self, *args, **kwargs):
         self._clear()
@@ -374,21 +495,45 @@ class FakeLogger(logging.Logger):
         self.thread_locals = None
         self.parent = None
 
+    store_in = {
+        logging.ERROR: 'error',
+        logging.WARNING: 'warning',
+        logging.INFO: 'info',
+        logging.DEBUG: 'debug',
+        logging.CRITICAL: 'critical',
+        NOTICE: 'notice',
+    }
+
+    def warn(self, *args, **kwargs):
+        raise WARN_DEPRECATED("Deprecated Method warn use warning instead")
+
+    def notice(self, msg, *args, **kwargs):
+        """
+        Convenience function for syslog priority LOG_NOTICE. The python
+        logging lvl is set to 25, just above info.  SysLogHandler is
+        monkey patched to map this log lvl to the LOG_NOTICE syslog
+        priority.
+        """
+        self.log(NOTICE, msg, *args, **kwargs)
+
+    def _log(self, level, msg, *args, **kwargs):
+        store_name = self.store_in[level]
+        cargs = [msg]
+        if any(args):
+            cargs.extend(args)
+        captured = dict(kwargs)
+        if 'exc_info' in kwargs and \
+                not isinstance(kwargs['exc_info'], tuple):
+            captured['exc_info'] = sys.exc_info()
+        self.log_dict[store_name].append((tuple(cargs), captured))
+        super(FakeLogger, self)._log(level, msg, *args, **kwargs)
+
     def _clear(self):
         self.log_dict = defaultdict(list)
         self.lines_dict = {'critical': [], 'error': [], 'info': [],
-                           'warning': [], 'debug': []}
+                           'warning': [], 'debug': [], 'notice': []}
 
-    def _store_in(store_name):
-        def stub_fn(self, *args, **kwargs):
-            self.log_dict[store_name].append((args, kwargs))
-        return stub_fn
-
-    def _store_and_log_in(store_name, level):
-        def stub_fn(self, *args, **kwargs):
-            self.log_dict[store_name].append((args, kwargs))
-            self._log(level, args[0], args[1:], **kwargs)
-        return stub_fn
+    clear = _clear  # this is a public interface
 
     def get_lines_for_level(self, level):
         if level not in self.lines_dict:
@@ -402,16 +547,10 @@ class FakeLogger(logging.Logger):
         return dict((level, msgs) for level, msgs in self.lines_dict.items()
                     if len(msgs) > 0)
 
-    error = _store_and_log_in('error', logging.ERROR)
-    info = _store_and_log_in('info', logging.INFO)
-    warning = _store_and_log_in('warning', logging.WARNING)
-    warn = _store_and_log_in('warning', logging.WARNING)
-    debug = _store_and_log_in('debug', logging.DEBUG)
-
-    def exception(self, *args, **kwargs):
-        self.log_dict['exception'].append((args, kwargs,
-                                           str(sys.exc_info()[1])))
-        print 'FakeLogger Exception: %s' % self.log_dict
+    def _store_in(store_name):
+        def stub_fn(self, *args, **kwargs):
+            self.log_dict[store_name].append((args, kwargs))
+        return stub_fn
 
     # mock out the StatsD logging methods:
     update_stats = _store_in('update_stats')
@@ -459,8 +598,8 @@ class FakeLogger(logging.Logger):
         try:
             line = record.getMessage()
         except TypeError:
-            print 'WARNING: unable to format log message %r %% %r' % (
-                record.msg, record.args)
+            print('WARNING: unable to format log message %r %% %r' % (
+                record.msg, record.args))
             raise
         self.lines_dict[record.levelname.lower()].append(line)
 
@@ -474,20 +613,27 @@ class FakeLogger(logging.Logger):
         pass
 
 
+class DebugSwiftLogFormatter(utils.SwiftLogFormatter):
+
+    def format(self, record):
+        msg = super(DebugSwiftLogFormatter, self).format(record)
+        return msg.replace('#012', '\n')
+
+
 class DebugLogger(FakeLogger):
     """A simple stdout logging version of FakeLogger"""
 
     def __init__(self, *args, **kwargs):
         FakeLogger.__init__(self, *args, **kwargs)
-        self.formatter = logging.Formatter(
+        self.formatter = DebugSwiftLogFormatter(
             "%(server)s %(levelname)s: %(message)s")
 
     def handle(self, record):
         self._handle(record)
-        print self.formatter.format(record)
+        print(self.formatter.format(record))
 
 
-class DebugLogAdapter(LogAdapter):
+class DebugLogAdapter(utils.LogAdapter):
 
     def _send_to_logger(name):
         def stub_fn(self, *args, **kwargs):
@@ -529,7 +675,8 @@ def fake_syslog_handler():
     logging.handlers.SysLogHandler = FakeLogger
 
 
-if config_true_value(get_config('unit_test').get('fake_syslog', 'False')):
+if utils.config_true_value(
+        get_config('unit_test').get('fake_syslog', 'False')):
     fake_syslog_handler()
 
 
@@ -602,32 +749,114 @@ def mock(update):
             delattr(module, attr)
 
 
+class FakeStatus(object):
+    """
+    This will work with our fake_http_connect, if you hand in one of these
+    instead of a status int or status int tuple to the "codes" iter you can
+    add some eventlet sleep to the expect and response stages of the
+    connection.
+    """
+
+    def __init__(self, status, expect_sleep=None, response_sleep=None):
+        """
+        :param status: the response status int, or a tuple of
+                       ([expect_status, ...], response_status)
+        :param expect_sleep: float, time to eventlet sleep during expect, can
+                             be a iter of floats
+        :param response_sleep: float, time to eventlet sleep during response
+        """
+        # connect exception
+        if isinstance(status, (Exception, eventlet.Timeout)):
+            raise status
+        if isinstance(status, tuple):
+            self.expect_status = list(status[:-1])
+            self.status = status[-1]
+            self.explicit_expect_list = True
+        else:
+            self.expect_status, self.status = ([], status)
+            self.explicit_expect_list = False
+        if not self.expect_status:
+            # when a swift backend service returns a status before reading
+            # from the body (mostly an error response) eventlet.wsgi will
+            # respond with that status line immediately instead of 100
+            # Continue, even if the client sent the Expect 100 header.
+            # BufferedHttp and the proxy both see these error statuses
+            # when they call getexpect, so our FakeConn tries to act like
+            # our backend services and return certain types of responses
+            # as expect statuses just like a real backend server would do.
+            if self.status in (507, 412, 409):
+                self.expect_status = [status]
+            else:
+                self.expect_status = [100, 100]
+
+        # setup sleep attributes
+        if not isinstance(expect_sleep, (list, tuple)):
+            expect_sleep = [expect_sleep] * len(self.expect_status)
+        self.expect_sleep_list = list(expect_sleep)
+        while len(self.expect_sleep_list) < len(self.expect_status):
+            self.expect_sleep_list.append(None)
+        self.response_sleep = response_sleep
+
+    def get_response_status(self):
+        if self.response_sleep is not None:
+            eventlet.sleep(self.response_sleep)
+        if self.expect_status and self.explicit_expect_list:
+            raise Exception('Test did not consume all fake '
+                            'expect status: %r' % (self.expect_status,))
+        if isinstance(self.status, (Exception, eventlet.Timeout)):
+            raise self.status
+        return self.status
+
+    def get_expect_status(self):
+        expect_sleep = self.expect_sleep_list.pop(0)
+        if expect_sleep is not None:
+            eventlet.sleep(expect_sleep)
+        expect_status = self.expect_status.pop(0)
+        if isinstance(expect_status, (Exception, eventlet.Timeout)):
+            raise expect_status
+        return expect_status
+
+
+class SlowBody(object):
+    """
+    This will work with our fake_http_connect, if you hand in these
+    instead of strings it will make reads take longer by the given
+    amount.  It should be a little bit easier to extend than the
+    current slow kwarg - which inserts whitespace in the response.
+    Also it should be easy to detect if you have one of these (or a
+    subclass) for the body inside of FakeConn if we wanted to do
+    something smarter than just duck-type the str/buffer api
+    enough to get by.
+    """
+
+    def __init__(self, body, slowness):
+        self.body = body
+        self.slowness = slowness
+
+    def slowdown(self):
+        eventlet.sleep(self.slowness)
+
+    def __getitem__(self, s):
+        return SlowBody(self.body[s], self.slowness)
+
+    def __len__(self):
+        return len(self.body)
+
+    def __radd__(self, other):
+        self.slowdown()
+        return other + self.body
+
+
 def fake_http_connect(*code_iter, **kwargs):
 
     class FakeConn(object):
 
         def __init__(self, status, etag=None, body='', timestamp='1',
-                     headers=None):
-            # connect exception
-            if isinstance(status, (Exception, Timeout)):
-                raise status
-            if isinstance(status, tuple):
-                self.expect_status, self.status = status
-            else:
-                self.expect_status, self.status = (None, status)
-            if not self.expect_status:
-                # when a swift backend service returns a status before reading
-                # from the body (mostly an error response) eventlet.wsgi will
-                # respond with that status line immediately instead of 100
-                # Continue, even if the client sent the Expect 100 header.
-                # BufferedHttp and the proxy both see these error statuses
-                # when they call getexpect, so our FakeConn tries to act like
-                # our backend services and return certain types of responses
-                # as expect statuses just like a real backend server would do.
-                if self.status in (507, 412, 409):
-                    self.expect_status = status
-                else:
-                    self.expect_status = 100
+                     headers=None, expect_headers=None, connection_id=None,
+                     give_send=None, give_expect=None):
+            if not isinstance(status, FakeStatus):
+                status = FakeStatus(status)
+            self._status = status
             self.reason = 'Fake'
             self.host = '1.2.3.4'
             self.port = '1234'
@@ -636,32 +865,43 @@ def fake_http_connect(*code_iter, **kwargs):
             self.etag = etag
             self.body = body
             self.headers = headers or {}
+            self.expect_headers = expect_headers or {}
             self.timestamp = timestamp
+            self.connection_id = connection_id
+            self.give_send = give_send
+            self.give_expect = give_expect
+            self.closed = False
             if 'slow' in kwargs and isinstance(kwargs['slow'], list):
                 try:
                     self._next_sleep = kwargs['slow'].pop(0)
                 except IndexError:
                     self._next_sleep = None
+            # be nice to trixy bits with node_iter's
+            eventlet.sleep()
 
         def getresponse(self):
-            if isinstance(self.status, (Exception, Timeout)):
-                raise self.status
             exc = kwargs.get('raise_exc')
             if exc:
-                if isinstance(exc, (Exception, Timeout)):
+                if isinstance(exc, (Exception, eventlet.Timeout)):
                     raise exc
                 raise Exception('test')
             if kwargs.get('raise_timeout_exc'):
-                raise Timeout()
+                raise eventlet.Timeout()
+            self.status = self._status.get_response_status()
             return self
 
         def getexpect(self):
-            if isinstance(self.expect_status, (Exception, Timeout)):
-                raise self.expect_status
-            headers = {}
-            if self.expect_status == 409:
+            if self.give_expect:
+                self.give_expect(self)
+            expect_status = self._status.get_expect_status()
+            headers = dict(self.expect_headers)
+            if expect_status == 409:
                 headers['X-Backend-Timestamp'] = self.timestamp
-            return FakeConn(self.expect_status, headers=headers)
+            response = FakeConn(expect_status,
+                                timestamp=self.timestamp,
+                                headers=headers)
+            response.status = expect_status
+            return response
 
         def getheaders(self):
             etag = self.etag
@@ -671,22 +911,25 @@ def fake_http_connect(*code_iter, **kwargs):
                 else:
                     etag = '"68b329da9893e34099c7d8ad5cb9c940"'
 
-            headers = {'content-length': len(self.body),
-                       'content-type': 'x-application/test',
-                       'x-timestamp': self.timestamp,
-                       'x-backend-timestamp': self.timestamp,
-                       'last-modified': self.timestamp,
-                       'x-object-meta-test': 'testing',
-                       'x-delete-at': '9876543210',
-                       'etag': etag,
-                       'x-works': 'yes'}
+            headers = HeaderKeyDict({
+                'content-length': len(self.body),
+                'content-type': 'x-application/test',
+                'x-timestamp': self.timestamp,
+                'x-backend-timestamp': self.timestamp,
+                'last-modified': self.timestamp,
+                'x-object-meta-test': 'testing',
+                'x-delete-at': '9876543210',
+                'etag': etag,
+                'x-works': 'yes',
+            })
             if self.status // 100 == 2:
                 headers['x-account-container-count'] = \
                     kwargs.get('count', 12345)
             if not self.timestamp:
-                del headers['x-timestamp']
+                # when timestamp is None, HeaderKeyDict raises KeyError
+                headers.pop('x-timestamp', None)
             try:
-                if container_ts_iter.next() is False:
+                if next(container_ts_iter) is False:
                     headers['x-container-timestamp'] = '1'
             except StopIteration:
                 pass
@@ -711,34 +954,45 @@ def fake_http_connect(*code_iter, **kwargs):
             if am_slow:
                 if self.sent < 4:
                     self.sent += 1
-                    sleep(value)
+                    eventlet.sleep(value)
                     return ' '
             rv = self.body[:amt]
             self.body = self.body[amt:]
             return rv
 
-        def send(self, amt=None):
+        def send(self, data=None):
+            if self.give_send:
+                self.give_send(self, data)
             am_slow, value = self.get_slow()
             if am_slow:
                 if self.received < 4:
                     self.received += 1
-                    sleep(value)
+                    eventlet.sleep(value)
 
         def getheader(self, name, default=None):
-            return dict(self.getheaders()).get(name.lower(), default)
+            return HeaderKeyDict(self.getheaders()).get(name, default)
+
+        def close(self):
+            self.closed = True
 
     timestamps_iter = iter(kwargs.get('timestamps') or ['1'] * len(code_iter))
     etag_iter = iter(kwargs.get('etags') or [None] * len(code_iter))
-    if isinstance(kwargs.get('headers'), list):
+    if isinstance(kwargs.get('headers'), (list, tuple)):
         headers_iter = iter(kwargs['headers'])
     else:
         headers_iter = iter([kwargs.get('headers', {})] * len(code_iter))
+    if isinstance(kwargs.get('expect_headers'), (list, tuple)):
+        expect_headers_iter = iter(kwargs['expect_headers'])
+    else:
+        expect_headers_iter = iter([kwargs.get('expect_headers', {})] *
+                                   len(code_iter))
 
     x = kwargs.get('missing_container', [False] * len(code_iter))
     if not isinstance(x, (tuple, list)):
         x = [x] * len(code_iter)
     container_ts_iter = iter(x)
     code_iter = iter(code_iter)
+    conn_id_and_code_iter = enumerate(code_iter)
     static_body = kwargs.get('body', None)
     body_iter = kwargs.get('body_iter', None)
     if body_iter:
@@ -746,27 +1000,34 @@ def fake_http_connect(*code_iter, **kwargs):
 
     def connect(*args, **ckwargs):
         if kwargs.get('slow_connect', False):
-            sleep(0.1)
+            eventlet.sleep(0.1)
         if 'give_content_type' in kwargs:
             if len(args) >= 7 and 'Content-Type' in args[6]:
                 kwargs['give_content_type'](args[6]['Content-Type'])
             else:
                 kwargs['give_content_type']('')
+        i, status = next(conn_id_and_code_iter)
         if 'give_connect' in kwargs:
-            kwargs['give_connect'](*args, **ckwargs)
-        status = code_iter.next()
-        etag = etag_iter.next()
-        headers = headers_iter.next()
-        timestamp = timestamps_iter.next()
+            give_conn_fn = kwargs['give_connect']
+            argspec = inspect.getargspec(give_conn_fn)
+            if argspec.keywords or 'connection_id' in argspec.args:
+                ckwargs['connection_id'] = i
+            give_conn_fn(*args, **ckwargs)
+        etag = next(etag_iter)
+        headers = next(headers_iter)
+        expect_headers = next(expect_headers_iter)
+        timestamp = next(timestamps_iter)
 
         if status <= 0:
             raise HTTPException()
         if body_iter is None:
             body = static_body or ''
         else:
-            body = body_iter.next()
+            body = next(body_iter)
         return FakeConn(status, etag, body=body, timestamp=timestamp,
-                        headers=headers)
+                        headers=headers, expect_headers=expect_headers,
+                        connection_id=i, give_send=kwargs.get('give_send'),
+                        give_expect=kwargs.get('give_expect'))
 
     connect.code_iter = code_iter
 
@@ -775,7 +1036,46 @@ def fake_http_connect(*code_iter, **kwargs):
 
 @contextmanager
 def mocked_http_conn(*args, **kwargs):
+    requests = []
+
+    def capture_requests(ip, port, method, path, headers, qs, ssl):
+        req = {
+            'ip': ip,
+            'port': port,
+            'method': method,
+            'path': path,
+            'headers': headers,
+            'qs': qs,
+            'ssl': ssl,
+        }
+        requests.append(req)
+    kwargs.setdefault('give_connect', capture_requests)
     fake_conn = fake_http_connect(*args, **kwargs)
+    fake_conn.requests = requests
     with mocklib.patch('swift.common.bufferedhttp.http_connect_raw',
                        new=fake_conn):
         yield fake_conn
+        left_over_status = list(fake_conn.code_iter)
+        if left_over_status:
+            raise AssertionError('left over status %r' % left_over_status)
+
+
+def make_timestamp_iter():
+    return iter(Timestamp(t) for t in itertools.count(int(time.time())))
+
+
+class Timeout(object):
+    def __init__(self, seconds):
+        self.seconds = seconds
+
+    def __enter__(self):
+        signal.signal(signal.SIGALRM, self._exit)
+        signal.alarm(self.seconds)
+
+    def __exit__(self, type, value, traceback):
+        signal.alarm(0)
+
+    def _exit(self, signum, frame):
+        class TimeoutException(Exception):
+            pass
+        raise TimeoutException
