@@ -13,24 +13,26 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from contextlib import nested
 import json
 import mock
 import os
-import random
 import re
-import string
 import tempfile
 import time
 import unittest
-import urlparse
+import shutil
+import sys
 
-from eventlet.green import urllib2
+from eventlet.green import urllib2, socket
 from six import StringIO
+from six.moves import urllib
 
 from swift.cli import recon
 from swift.common import utils
 from swift.common.ring import builder
+from swift.common.ring import utils as ring_utils
+from swift.common.storage_policy import StoragePolicy, POLICIES
+from test.unit import patch_policies
 
 
 class TestHelpers(unittest.TestCase):
@@ -86,6 +88,15 @@ class TestScout(unittest.TestCase):
         self.assertTrue(isinstance(content, urllib2.HTTPError))
         self.assertEqual(status, 404)
 
+    @mock.patch('eventlet.green.urllib2.urlopen')
+    def test_scout_socket_timeout(self, mock_urlopen):
+        mock_urlopen.side_effect = socket.timeout("timeout")
+        url, content, status, ts_start, ts_end = self.scout_instance.scout(
+            ("127.0.0.1", "8080"))
+        self.assertTrue(isinstance(content, socket.timeout))
+        self.assertEqual(url, self.url)
+        self.assertEqual(status, -1)
+
     @mock.patch('eventlet.green.urllib2.urlopen')
     def test_scout_server_type_ok(self, mock_urlopen):
         def getheader(name):
@@ -117,23 +128,60 @@ class TestScout(unittest.TestCase):
         self.assertTrue(isinstance(content, urllib2.HTTPError))
         self.assertEqual(status, 404)
 
+    @mock.patch('eventlet.green.urllib2.urlopen')
+    def test_scout_server_type_socket_timeout(self, mock_urlopen):
+        mock_urlopen.side_effect = socket.timeout("timeout")
+        url, content, status = self.scout_instance.scout_server_type(
+            ("127.0.0.1", "8080"))
+        self.assertTrue(isinstance(content, socket.timeout))
+        self.assertEqual(url, self.server_type_url)
+        self.assertEqual(status, -1)
 
+
+@patch_policies
 class TestRecon(unittest.TestCase):
     def setUp(self, *_args, **_kwargs):
         self.recon_instance = recon.SwiftRecon()
-        self.swift_dir = tempfile.gettempdir()
-        self.ring_name = "test_object_%s" % (
-            ''.join(random.choice(string.digits) for x in range(6)))
-        self.tmpfile_name = "%s/%s.ring.gz" % (self.swift_dir, self.ring_name)
+        self.swift_dir = tempfile.mkdtemp()
+        self.ring_name = POLICIES.legacy.ring_name
+        self.tmpfile_name = os.path.join(
+            self.swift_dir, self.ring_name + '.ring.gz')
+        self.ring_name2 = POLICIES[1].ring_name
+        self.tmpfile_name2 = os.path.join(
+            self.swift_dir, self.ring_name2 + '.ring.gz')
 
         utils.HASH_PATH_SUFFIX = 'endcap'
         utils.HASH_PATH_PREFIX = 'startcap'
 
     def tearDown(self, *_args, **_kwargs):
-        try:
-            os.remove(self.tmpfile_name)
-        except OSError:
-            pass
+        shutil.rmtree(self.swift_dir, ignore_errors=True)
+
+    def _make_object_rings(self):
+        ringbuilder = builder.RingBuilder(2, 3, 1)
+        devs = [
+            'r0z0-127.0.0.1:10000/sda1',
+            'r0z1-127.0.0.1:10001/sda1',
+            'r1z0-127.0.0.1:10002/sda1',
+            'r1z1-127.0.0.1:10003/sda1',
+        ]
+        for raw_dev_str in devs:
+            dev = ring_utils.parse_add_value(raw_dev_str)
+            dev['weight'] = 1.0
+            ringbuilder.add_dev(dev)
+        ringbuilder.rebalance()
+        ringbuilder.get_ring().save(self.tmpfile_name)
+
+        ringbuilder = builder.RingBuilder(2, 2, 1)
+        devs = [
+            'r0z0-127.0.0.1:10000/sda1',
+            'r0z1-127.0.0.2:10004/sda1',
+        ]
+        for raw_dev_str in devs:
+            dev = ring_utils.parse_add_value(raw_dev_str)
+            dev['weight'] = 1.0
+            ringbuilder.add_dev(dev)
+        ringbuilder.rebalance()
+        ringbuilder.get_ring().save(self.tmpfile_name2)
 
     def test_gen_stats(self):
         stats = self.recon_instance._gen_stats((1, 4, 10, None), 'Sample')
@@ -147,59 +195,68 @@ class TestRecon(unittest.TestCase):
         self.assertEqual(stats.get('perc_none'), 25.0)
 
     def test_ptime(self):
-        with mock.patch('time.localtime') as mock_localtime:
-            mock_localtime.return_value = time.struct_time(
+        with mock.patch('time.gmtime') as mock_gmtime:
+            mock_gmtime.return_value = time.struct_time(
                 (2013, 12, 17, 10, 0, 0, 1, 351, 0))
 
             timestamp = self.recon_instance._ptime(1387274400)
             self.assertEqual(timestamp, "2013-12-17 10:00:00")
-            mock_localtime.assert_called_with(1387274400)
+            mock_gmtime.assert_called_with(1387274400)
 
             timestamp2 = self.recon_instance._ptime()
             self.assertEqual(timestamp2, "2013-12-17 10:00:00")
-            mock_localtime.assert_called_with()
+            mock_gmtime.assert_called_with()
 
-    def test_get_devices(self):
-        ringbuilder = builder.RingBuilder(2, 3, 1)
-        ringbuilder.add_dev({'id': 0, 'zone': 0, 'weight': 1,
-                             'ip': '127.0.0.1', 'port': 10000,
-                             'device': 'sda1', 'region': 0})
-        ringbuilder.add_dev({'id': 1, 'zone': 1, 'weight': 1,
-                             'ip': '127.0.0.1', 'port': 10001,
-                             'device': 'sda1', 'region': 0})
-        ringbuilder.add_dev({'id': 2, 'zone': 0, 'weight': 1,
-                             'ip': '127.0.0.1', 'port': 10002,
-                             'device': 'sda1', 'region': 1})
-        ringbuilder.add_dev({'id': 3, 'zone': 1, 'weight': 1,
-                             'ip': '127.0.0.1', 'port': 10003,
-                             'device': 'sda1', 'region': 1})
-        ringbuilder.rebalance()
-        ringbuilder.get_ring().save(self.tmpfile_name)
+    def test_get_hosts(self):
+        self._make_object_rings()
 
-        ips = self.recon_instance.get_devices(
-            None, None, self.swift_dir, self.ring_name)
+        ips = self.recon_instance.get_hosts(
+            None, None, self.swift_dir, [self.ring_name])
         self.assertEqual(
             set([('127.0.0.1', 10000), ('127.0.0.1', 10001),
                  ('127.0.0.1', 10002), ('127.0.0.1', 10003)]), ips)
 
-        ips = self.recon_instance.get_devices(
-            0, None, self.swift_dir, self.ring_name)
+        ips = self.recon_instance.get_hosts(
+            0, None, self.swift_dir, [self.ring_name])
         self.assertEqual(
             set([('127.0.0.1', 10000), ('127.0.0.1', 10001)]), ips)
 
-        ips = self.recon_instance.get_devices(
-            1, None, self.swift_dir, self.ring_name)
+        ips = self.recon_instance.get_hosts(
+            1, None, self.swift_dir, [self.ring_name])
         self.assertEqual(
             set([('127.0.0.1', 10002), ('127.0.0.1', 10003)]), ips)
 
-        ips = self.recon_instance.get_devices(
-            0, 0, self.swift_dir, self.ring_name)
+        ips = self.recon_instance.get_hosts(
+            0, 0, self.swift_dir, [self.ring_name])
         self.assertEqual(set([('127.0.0.1', 10000)]), ips)
 
-        ips = self.recon_instance.get_devices(
-            1, 1, self.swift_dir, self.ring_name)
+        ips = self.recon_instance.get_hosts(
+            1, 1, self.swift_dir, [self.ring_name])
         self.assertEqual(set([('127.0.0.1', 10003)]), ips)
 
+        ips = self.recon_instance.get_hosts(
+            None, None, self.swift_dir, [self.ring_name, self.ring_name2])
+        self.assertEqual(
+            set([('127.0.0.1', 10000), ('127.0.0.1', 10001),
+                 ('127.0.0.1', 10002), ('127.0.0.1', 10003),
+                 ('127.0.0.2', 10004)]), ips)
+
+        ips = self.recon_instance.get_hosts(
+            0, None, self.swift_dir, [self.ring_name, self.ring_name2])
+        self.assertEqual(
+            set([('127.0.0.1', 10000), ('127.0.0.1', 10001),
+                 ('127.0.0.2', 10004)]), ips)
+
+        ips = self.recon_instance.get_hosts(
+            1, None, self.swift_dir, [self.ring_name, self.ring_name2])
+        self.assertEqual(
+            set([('127.0.0.1', 10002), ('127.0.0.1', 10003)]), ips)
+
+        ips = self.recon_instance.get_hosts(
+            0, 1, self.swift_dir, [self.ring_name, self.ring_name2])
+        self.assertEqual(set([('127.0.0.1', 10001),
+                              ('127.0.0.2', 10004)]), ips)
+
     def test_get_ringmd5(self):
         for server_type in ('account', 'container', 'object', 'object-1'):
             ring_name = '%s.ring.gz' % server_type
@@ -222,11 +279,8 @@ class TestRecon(unittest.TestCase):
             mock_scout.return_value = scout_instance
             stdout = StringIO()
             mock_hash = mock.MagicMock()
-            patches = [
-                mock.patch('sys.stdout', new=stdout),
-                mock.patch('swift.cli.recon.md5', new=mock_hash),
-            ]
-            with nested(*patches):
+            with mock.patch('sys.stdout', new=stdout), \
+                    mock.patch('swift.cli.recon.md5', new=mock_hash):
                 mock_hash.return_value.hexdigest.return_value = \
                     empty_file_hash
                 self.recon_instance.get_ringmd5(hosts, self.swift_dir)
@@ -277,11 +331,9 @@ class TestRecon(unittest.TestCase):
             return url, response, status, 0, 0
 
         stdout = StringIO()
-        patches = [
-            mock.patch('swift.cli.recon.Scout.scout', mock_scout_quarantine),
-            mock.patch('sys.stdout', new=stdout),
-        ]
-        with nested(*patches):
+        with mock.patch('swift.cli.recon.Scout.scout',
+                        mock_scout_quarantine), \
+                mock.patch('sys.stdout', new=stdout):
             self.recon_instance.quarantine_check(hosts)
 
         output = stdout.getvalue()
@@ -296,6 +348,81 @@ class TestRecon(unittest.TestCase):
                                  % ex)
         self.assertFalse(expected)
 
+    def test_async_check(self):
+        hosts = [('127.0.0.1', 6011), ('127.0.0.1', 6021),
+                 ('127.0.0.1', 6031), ('127.0.0.1', 6041)]
+        # sample json response from http://<host>:<port>/recon/async
+        responses = {6011: {'async_pending': 15},
+                     6021: {'async_pending': 0},
+                     6031: {'async_pending': 257},
+                     6041: {'async_pending': 56}}
+        # <low> <high> <avg> <total> <Failed> <no_result> <reported>
+        expected = (0, 257, 82.0, 328, 0.0, 0, 4)
+
+        def mock_scout_async(app, host):
+            url = 'http://%s:%s/recon/async' % host
+            response = responses[host[1]]
+            status = 200
+            return url, response, status, 0, 0
+
+        stdout = StringIO()
+        with mock.patch('swift.cli.recon.Scout.scout',
+                        mock_scout_async), \
+                mock.patch('sys.stdout', new=stdout):
+            self.recon_instance.async_check(hosts)
+
+        output = stdout.getvalue()
+        r = re.compile("\[async_pending(.*)\](.*)")
+        lines = output.splitlines()
+        self.assertTrue(lines)
+        for line in lines:
+            m = r.match(line)
+            if m:
+                self.assertEqual(m.group(2),
+                                 " low: %s, high: %s, avg: %s, total: %s,"
+                                 " Failed: %s%%, no_result: %s, reported: %s"
+                                 % expected)
+                break
+        else:
+            self.fail('The expected line is not found')
+
+    def test_umount_check(self):
+        hosts = [('127.0.0.1', 6010), ('127.0.0.1', 6020),
+                 ('127.0.0.1', 6030), ('127.0.0.1', 6040)]
+        # sample json response from http://<host>:<port>/recon/unmounted
+        responses = {6010: [{'device': 'sdb1', 'mounted': False}],
+                     6020: [{'device': 'sdb2', 'mounted': False}],
+                     6030: [{'device': 'sdb3', 'mounted': False}],
+                     6040: [{'device': 'sdb4', 'mounted': 'bad'}]}
+
+        expected = ['Not mounted: sdb1 on 127.0.0.1:6010',
+                    'Not mounted: sdb2 on 127.0.0.1:6020',
+                    'Not mounted: sdb3 on 127.0.0.1:6030',
+                    'Device errors: sdb4 on 127.0.0.1:6040']
+
+        def mock_scout_umount(app, host):
+            url = 'http://%s:%s/recon/unmounted' % host
+            response = responses[host[1]]
+            status = 200
+            return url, response, status, 0, 0
+
+        stdout = StringIO()
+        with mock.patch('swift.cli.recon.Scout.scout',
+                        mock_scout_umount), \
+                mock.patch('sys.stdout', new=stdout):
+            self.recon_instance.umount_check(hosts)
+
+        output = stdout.getvalue()
+        r = re.compile("\Not mounted:|Device errors: .*")
+        lines = output.splitlines()
+        self.assertTrue(lines)
+        for line in lines:
+            m = r.match(line)
+            if m:
+                self.assertIn(line, expected)
+                expected.remove(line)
+        self.assertFalse(expected)
+
     def test_drive_audit_check(self):
         hosts = [('127.0.0.1', 6010), ('127.0.0.1', 6020),
                  ('127.0.0.1', 6030), ('127.0.0.1', 6040)]
@@ -314,11 +441,9 @@ class TestRecon(unittest.TestCase):
             return url, response, status, 0, 0
 
         stdout = StringIO()
-        patches = [
-            mock.patch('swift.cli.recon.Scout.scout', mock_scout_driveaudit),
-            mock.patch('sys.stdout', new=stdout),
-        ]
-        with nested(*patches):
+        with mock.patch('swift.cli.recon.Scout.scout',
+                        mock_scout_driveaudit), \
+                mock.patch('sys.stdout', new=stdout):
             self.recon_instance.driveaudit_check(hosts)
 
         output = stdout.getvalue()
@@ -333,6 +458,89 @@ class TestRecon(unittest.TestCase):
                                  " Failed: %s%%, no_result: %s, reported: %s"
                                  % expected)
 
+    def test_get_ring_names(self):
+        self.recon_instance.server_type = 'not-object'
+        self.assertEqual(self.recon_instance._get_ring_names(), ['not-object'])
+
+        self.recon_instance.server_type = 'object'
+
+        with patch_policies([StoragePolicy(0, 'zero', is_default=True)]):
+            self.assertEqual(self.recon_instance._get_ring_names(),
+                             ['object'])
+
+        with patch_policies([StoragePolicy(0, 'zero', is_default=True),
+                             StoragePolicy(1, 'one')]):
+            self.assertEqual(self.recon_instance._get_ring_names(),
+                             ['object', 'object-1'])
+            self.assertEqual(self.recon_instance._get_ring_names('0'),
+                             ['object'])
+            self.assertEqual(self.recon_instance._get_ring_names('zero'),
+                             ['object'])
+            self.assertEqual(self.recon_instance._get_ring_names('1'),
+                             ['object-1'])
+            self.assertEqual(self.recon_instance._get_ring_names('one'),
+                             ['object-1'])
+
+            self.assertEqual(self.recon_instance._get_ring_names('3'), [])
+            self.assertEqual(self.recon_instance._get_ring_names('wrong'),
+                             [])
+
+    def test_main_object_hosts_default_all_policies(self):
+        self._make_object_rings()
+        discovered_hosts = set()
+
+        def server_type_check(hosts):
+            for h in hosts:
+                discovered_hosts.add(h)
+
+        self.recon_instance.server_type_check = server_type_check
+        with mock.patch.object(sys, 'argv', [
+                "prog", "object", "--swiftdir=%s" % self.swift_dir,
+                "--validate-servers"]):
+            self.recon_instance.main()
+
+        expected = set([
+            ('127.0.0.1', 10000),
+            ('127.0.0.1', 10001),
+            ('127.0.0.1', 10002),
+            ('127.0.0.1', 10003),
+            ('127.0.0.2', 10004),
+        ])
+
+        self.assertEqual(expected, discovered_hosts)
+
+    def test_main_object_hosts_default_unu(self):
+        self._make_object_rings()
+        discovered_hosts = set()
+
+        def server_type_check(hosts):
+            for h in hosts:
+                discovered_hosts.add(h)
+
+        self.recon_instance.server_type_check = server_type_check
+
+        with mock.patch.object(sys, 'argv', [
+                "prog", "object", "--swiftdir=%s" % self.swift_dir,
+                "--validate-servers", '--policy=unu']):
+
+            self.recon_instance.main()
+
+        expected = set([
+            ('127.0.0.1', 10000),
+            ('127.0.0.2', 10004),
+        ])
+        self.assertEqual(expected, discovered_hosts)
+
+    def test_main_object_hosts_default_invalid(self):
+        self._make_object_rings()
+        stdout = StringIO()
+        with mock.patch.object(sys, 'argv', [
+                "prog", "object", "--swiftdir=%s" % self.swift_dir,
+                "--validate-servers", '--policy=invalid']),\
+                mock.patch('sys.stdout', stdout):
+            self.assertRaises(SystemExit, recon.main)
+            self.assertIn('Invalid Storage Policy', stdout.getvalue())
+
 
 class TestReconCommands(unittest.TestCase):
     def setUp(self):
@@ -342,7 +550,7 @@ class TestReconCommands(unittest.TestCase):
     def mock_responses(self, resps):
 
         def fake_urlopen(url, timeout):
-            scheme, netloc, path, _, _, _ = urlparse.urlparse(url)
+            scheme, netloc, path, _, _, _ = urllib.parse.urlparse(url)
             self.assertEqual(scheme, 'http')  # can't handle anything else
             self.assertTrue(path.startswith('/recon/'))
 
@@ -376,19 +584,15 @@ class TestReconCommands(unittest.TestCase):
             return url, response, status
 
         stdout = StringIO()
-        patches = [
-            mock.patch('swift.cli.recon.Scout.scout_server_type',
-                       mock_scout_server_type),
-            mock.patch('sys.stdout', new=stdout),
-        ]
-
         res_object = 'Invalid: http://127.0.0.1:6010/ is object-server'
         res_container = 'Invalid: http://127.0.0.1:6011/ is container-server'
         res_account = 'Invalid: http://127.0.0.1:6012/ is account-server'
         valid = "1/1 hosts ok, 0 error[s] while checking hosts."
 
         # Test for object server type - default
-        with nested(*patches):
+        with mock.patch('swift.cli.recon.Scout.scout_server_type',
+                        mock_scout_server_type), \
+                mock.patch('sys.stdout', new=stdout):
             self.recon.server_type_check(hosts)
 
         output = stdout.getvalue()
@@ -397,7 +601,9 @@ class TestReconCommands(unittest.TestCase):
         stdout.truncate(0)
 
         # Test ok for object server type - default
-        with nested(*patches):
+        with mock.patch('swift.cli.recon.Scout.scout_server_type',
+                        mock_scout_server_type), \
+                mock.patch('sys.stdout', new=stdout):
             self.recon.server_type_check([hosts[0]])
 
         output = stdout.getvalue()
@@ -405,7 +611,9 @@ class TestReconCommands(unittest.TestCase):
         stdout.truncate(0)
 
         # Test for account server type
-        with nested(*patches):
+        with mock.patch('swift.cli.recon.Scout.scout_server_type',
+                        mock_scout_server_type), \
+                mock.patch('sys.stdout', new=stdout):
             self.recon.server_type = 'account'
             self.recon.server_type_check(hosts)
 
@@ -415,7 +623,9 @@ class TestReconCommands(unittest.TestCase):
         stdout.truncate(0)
 
         # Test ok for account server type
-        with nested(*patches):
+        with mock.patch('swift.cli.recon.Scout.scout_server_type',
+                        mock_scout_server_type), \
+                mock.patch('sys.stdout', new=stdout):
             self.recon.server_type = 'account'
             self.recon.server_type_check([hosts[2]])
 
@@ -424,7 +634,9 @@ class TestReconCommands(unittest.TestCase):
         stdout.truncate(0)
 
         # Test for container server type
-        with nested(*patches):
+        with mock.patch('swift.cli.recon.Scout.scout_server_type',
+                        mock_scout_server_type), \
+                mock.patch('sys.stdout', new=stdout):
             self.recon.server_type = 'container'
             self.recon.server_type_check(hosts)
 
@@ -434,7 +646,9 @@ class TestReconCommands(unittest.TestCase):
         stdout.truncate(0)
 
         # Test ok for container server type
-        with nested(*patches):
+        with mock.patch('swift.cli.recon.Scout.scout_server_type',
+                        mock_scout_server_type), \
+                mock.patch('sys.stdout', new=stdout):
             self.recon.server_type = 'container'
             self.recon.server_type_check([hosts[1]])
 
@@ -554,11 +768,11 @@ class TestReconCommands(unittest.TestCase):
             mock.call('=' * 79),
         ]
 
-        with mock.patch('__builtin__.print') as mock_print:
+        with mock.patch('six.moves.builtins.print') as mock_print:
             cli.disk_usage([('127.0.0.1', 6010)])
             mock_print.assert_has_calls(default_calls)
 
-        with mock.patch('__builtin__.print') as mock_print:
+        with mock.patch('six.moves.builtins.print') as mock_print:
             expected_calls = default_calls + [
                 mock.call('LOWEST 5'),
                 mock.call('85.00%  127.0.0.1       sdc1'),
@@ -568,7 +782,7 @@ class TestReconCommands(unittest.TestCase):
             cli.disk_usage([('127.0.0.1', 6010)], 0, 5)
             mock_print.assert_has_calls(expected_calls)
 
-        with mock.patch('__builtin__.print') as mock_print:
+        with mock.patch('six.moves.builtins.print') as mock_print:
             expected_calls = default_calls + [
                 mock.call('TOP 5'),
                 mock.call('90.00%  127.0.0.1       sdb1'),
@@ -578,7 +792,7 @@ class TestReconCommands(unittest.TestCase):
             cli.disk_usage([('127.0.0.1', 6010)], 5, 0)
             mock_print.assert_has_calls(expected_calls)
 
-    @mock.patch('__builtin__.print')
+    @mock.patch('six.moves.builtins.print')
     @mock.patch('time.time')
     def test_replication_check(self, mock_now, mock_print):
         now = 1430000000.0
@@ -633,7 +847,7 @@ class TestReconCommands(unittest.TestCase):
         # that is returned from the recon middleware, thus can't rely on it
         mock_print.assert_has_calls(default_calls, any_order=True)
 
-    @mock.patch('__builtin__.print')
+    @mock.patch('six.moves.builtins.print')
     @mock.patch('time.time')
     def test_load_check(self, mock_now, mock_print):
         now = 1430000000.0
@@ -672,7 +886,7 @@ class TestReconCommands(unittest.TestCase):
         # that is returned from the recon middleware, thus can't rely on it
         mock_print.assert_has_calls(default_calls, any_order=True)
 
-    @mock.patch('__builtin__.print')
+    @mock.patch('six.moves.builtins.print')
     @mock.patch('time.time')
     def test_time_check(self, mock_now, mock_print):
         now = 1430000000.0
@@ -704,7 +918,7 @@ class TestReconCommands(unittest.TestCase):
         # that is returned from the recon middleware, thus can't rely on it
         mock_print.assert_has_calls(default_calls, any_order=True)
 
-    @mock.patch('__builtin__.print')
+    @mock.patch('six.moves.builtins.print')
     @mock.patch('time.time')
     def test_time_check_mismatch(self, mock_now, mock_print):
         now = 1430000000.0
@@ -734,12 +948,32 @@ class TestReconCommands(unittest.TestCase):
             mock.call('1/2 hosts matched, 0 error[s] while checking hosts.'),
         ]
 
-        def mock_localtime(*args, **kwargs):
-            return time.gmtime(*args, **kwargs)
-
-        with mock.patch("time.localtime", mock_localtime):
-            cli.time_check([('127.0.0.1', 6010), ('127.0.0.1', 6020)])
+        cli.time_check([('127.0.0.1', 6010), ('127.0.0.1', 6020)])
 
         # We need any_order=True because the order of calls depends on the dict
         # that is returned from the recon middleware, thus can't rely on it
         mock_print.assert_has_calls(default_calls, any_order=True)
+
+    @mock.patch('six.moves.builtins.print')
+    @mock.patch('swift.cli.recon.SwiftRecon.get_hosts')
+    def test_multiple_server_types(self, mock_get_hosts, mock_print):
+        mock_get_hosts.return_value = set([('127.0.0.1', 10000)])
+
+        self.recon.object_auditor_check = mock.MagicMock()
+        self.recon.auditor_check = mock.MagicMock()
+
+        with mock.patch.object(
+                sys, 'argv',
+                ["prog", "account", "container", "object", "--auditor"]):
+            self.recon.main()
+        expected_calls = [
+            mock.call("--> Starting reconnaissance on 1 hosts (account)"),
+            mock.call("--> Starting reconnaissance on 1 hosts (container)"),
+            mock.call("--> Starting reconnaissance on 1 hosts (object)"),
+        ]
+        mock_print.assert_has_calls(expected_calls, any_order=True)
+
+        expected = mock.call(set([('127.0.0.1', 10000)]))
+        self.recon.object_auditor_check.assert_has_calls([expected])
+        # Two calls expected - one account, one container
+        self.recon.auditor_check.assert_has_calls([expected, expected])
