@@ -16,12 +16,14 @@
 
 import os
 import unittest
-from contextlib import nested
 from textwrap import dedent
 
 import mock
+import errno
+from swift.common.utils import Timestamp
 from test.unit import debug_logger
 from swift.container import sync
+from swift.common.db import DatabaseConnectionError
 from swift.common import utils
 from swift.common.wsgi import ConfigString
 from swift.common.exceptions import ClientException
@@ -48,6 +50,7 @@ class FakeContainerBroker(object):
     def __init__(self, path, metadata=None, info=None, deleted=False,
                  items_since=None):
         self.db_file = path
+        self.db_dir = os.path.dirname(path)
         self.metadata = metadata if metadata else {}
         self.info = info if info else {}
         self.deleted = deleted
@@ -55,6 +58,9 @@ class FakeContainerBroker(object):
         self.sync_point1 = -1
         self.sync_point2 = -1
 
+    def get_max_row(self):
+        return 1
+
     def get_info(self):
         return self.info
 
@@ -158,7 +164,6 @@ class TestContainerSync(unittest.TestCase):
         # interval sleep.
         time_calls = [0]
         sleep_calls = []
-        audit_location_generator_calls = [0]
 
         def fake_time():
             time_calls[0] += 1
@@ -177,48 +182,36 @@ class TestContainerSync(unittest.TestCase):
         def fake_sleep(amount):
             sleep_calls.append(amount)
 
-        def fake_audit_location_generator(*args, **kwargs):
-            audit_location_generator_calls[0] += 1
-            # Makes .container_sync() short-circuit
-            yield 'container.db', 'device', 'partition'
-            return
+        gen_func = ('swift.container.sync_store.'
+                    'ContainerSyncStore.synced_containers_generator')
+        with mock.patch('swift.container.sync.InternalClient'), \
+                mock.patch('swift.container.sync.time', fake_time), \
+                mock.patch('swift.container.sync.sleep', fake_sleep), \
+                mock.patch(gen_func) as fake_generator, \
+                mock.patch('swift.container.sync.ContainerBroker',
+                           lambda p: FakeContainerBroker(p, info={
+                               'account': 'a', 'container': 'c',
+                               'storage_policy_index': 0})):
+            fake_generator.side_effect = [iter(['container.db']),
+                                          iter(['container.db'])]
+            cs = sync.ContainerSync({}, container_ring=FakeRing())
+            try:
+                cs.run_forever()
+            except Exception as err:
+                if str(err) != 'we are now done':
+                    raise
 
-        orig_time = sync.time
-        orig_sleep = sync.sleep
-        orig_ContainerBroker = sync.ContainerBroker
-        orig_audit_location_generator = sync.audit_location_generator
-        try:
-            sync.ContainerBroker = lambda p: FakeContainerBroker(
-                p, info={'account': 'a', 'container': 'c',
-                         'storage_policy_index': 0})
-            sync.time = fake_time
-            sync.sleep = fake_sleep
-
-            with mock.patch('swift.container.sync.InternalClient'):
-                cs = sync.ContainerSync({}, container_ring=FakeRing())
-            sync.audit_location_generator = fake_audit_location_generator
-            cs.run_forever(1, 2, a=3, b=4, verbose=True)
-        except Exception as err:
-            if str(err) != 'we are now done':
-                raise
-        finally:
-            sync.time = orig_time
-            sync.sleep = orig_sleep
-            sync.audit_location_generator = orig_audit_location_generator
-            sync.ContainerBroker = orig_ContainerBroker
-
-        self.assertEqual(time_calls, [9])
-        self.assertEqual(len(sleep_calls), 2)
-        self.assertTrue(sleep_calls[0] <= cs.interval)
-        self.assertTrue(sleep_calls[1] == cs.interval - 1)
-        self.assertEqual(audit_location_generator_calls, [2])
-        self.assertEqual(cs.reported, 3602)
+            self.assertEqual(time_calls, [9])
+            self.assertEqual(len(sleep_calls), 2)
+            self.assertLessEqual(sleep_calls[0], cs.interval)
+            self.assertEqual(cs.interval - 1, sleep_calls[1])
+            self.assertEqual(2, fake_generator.call_count)
+            self.assertEqual(cs.reported, 3602)
 
     def test_run_once(self):
         # This runs runs_once with fakes twice, the first causing an interim
         # report, the second with no interim report.
         time_calls = [0]
-        audit_location_generator_calls = [0]
 
         def fake_time():
             time_calls[0] += 1
@@ -236,40 +229,31 @@ class TestContainerSync(unittest.TestCase):
                 raise Exception('we are now done')
             return returns[time_calls[0] - 1]
 
-        def fake_audit_location_generator(*args, **kwargs):
-            audit_location_generator_calls[0] += 1
-            # Makes .container_sync() short-circuit
-            yield 'container.db', 'device', 'partition'
-            return
+        gen_func = ('swift.container.sync_store.'
+                    'ContainerSyncStore.synced_containers_generator')
+        with mock.patch('swift.container.sync.InternalClient'), \
+                mock.patch('swift.container.sync.time', fake_time), \
+                mock.patch(gen_func) as fake_generator, \
+                mock.patch('swift.container.sync.ContainerBroker',
+                           lambda p: FakeContainerBroker(p, info={
+                               'account': 'a', 'container': 'c',
+                               'storage_policy_index': 0})):
+            fake_generator.side_effect = [iter(['container.db']),
+                                          iter(['container.db'])]
+            cs = sync.ContainerSync({}, container_ring=FakeRing())
+            try:
+                cs.run_once()
+                self.assertEqual(time_calls, [6])
+                self.assertEqual(1, fake_generator.call_count)
+                self.assertEqual(cs.reported, 3602)
+                cs.run_once()
+            except Exception as err:
+                if str(err) != 'we are now done':
+                    raise
 
-        orig_time = sync.time
-        orig_audit_location_generator = sync.audit_location_generator
-        orig_ContainerBroker = sync.ContainerBroker
-        try:
-            sync.ContainerBroker = lambda p: FakeContainerBroker(
-                p, info={'account': 'a', 'container': 'c',
-                         'storage_policy_index': 0})
-            sync.time = fake_time
-
-            with mock.patch('swift.container.sync.InternalClient'):
-                cs = sync.ContainerSync({}, container_ring=FakeRing())
-            sync.audit_location_generator = fake_audit_location_generator
-            cs.run_once(1, 2, a=3, b=4, verbose=True)
-            self.assertEqual(time_calls, [6])
-            self.assertEqual(audit_location_generator_calls, [1])
-            self.assertEqual(cs.reported, 3602)
-            cs.run_once()
-        except Exception as err:
-            if str(err) != 'we are now done':
-                raise
-        finally:
-            sync.time = orig_time
-            sync.audit_location_generator = orig_audit_location_generator
-            sync.ContainerBroker = orig_ContainerBroker
-
-        self.assertEqual(time_calls, [10])
-        self.assertEqual(audit_location_generator_calls, [2])
-        self.assertEqual(cs.reported, 3604)
+            self.assertEqual(time_calls, [10])
+            self.assertEqual(2, fake_generator.call_count)
+            self.assertEqual(cs.reported, 3604)
 
     def test_container_sync_not_db(self):
         cring = FakeRing()
@@ -281,8 +265,65 @@ class TestContainerSync(unittest.TestCase):
         cring = FakeRing()
         with mock.patch('swift.container.sync.InternalClient'):
             cs = sync.ContainerSync({}, container_ring=cring)
-        cs.container_sync('isa.db')
-        self.assertEqual(cs.container_failures, 1)
+
+        broker = 'swift.container.backend.ContainerBroker'
+        store = 'swift.container.sync_store.ContainerSyncStore'
+
+        # In this test we call the container_sync instance several
+        # times with a missing db in various combinations.
+        # Since we use the same ContainerSync instance for all tests
+        # its failures counter increases by one with each call.
+
+        # Test the case where get_info returns DatabaseConnectionError
+        # with DB does not exist, and we succeed in deleting it.
+        with mock.patch(broker + '.get_info') as fake_get_info:
+            with mock.patch(store + '.remove_synced_container') as fake_remove:
+                fake_get_info.side_effect = DatabaseConnectionError(
+                    'a',
+                    "DB doesn't exist")
+                cs.container_sync('isa.db')
+                self.assertEqual(cs.container_failures, 1)
+                self.assertEqual(cs.container_skips, 0)
+                self.assertEqual(1, fake_remove.call_count)
+                self.assertEqual('isa.db', fake_remove.call_args[0][0].db_file)
+
+        # Test the case where get_info returns DatabaseConnectionError
+        # with DB does not exist, and we fail to delete it.
+        with mock.patch(broker + '.get_info') as fake_get_info:
+            with mock.patch(store + '.remove_synced_container') as fake_remove:
+                fake_get_info.side_effect = DatabaseConnectionError(
+                    'a',
+                    "DB doesn't exist")
+                fake_remove.side_effect = OSError('1')
+                cs.container_sync('isa.db')
+                self.assertEqual(cs.container_failures, 2)
+                self.assertEqual(cs.container_skips, 0)
+                self.assertEqual(1, fake_remove.call_count)
+                self.assertEqual('isa.db', fake_remove.call_args[0][0].db_file)
+
+        # Test the case where get_info returns DatabaseConnectionError
+        # with DB does not exist, and it returns an error != ENOENT.
+        with mock.patch(broker + '.get_info') as fake_get_info:
+            with mock.patch(store + '.remove_synced_container') as fake_remove:
+                fake_get_info.side_effect = DatabaseConnectionError(
+                    'a',
+                    "DB doesn't exist")
+                fake_remove.side_effect = OSError(errno.EPERM, 'a')
+                cs.container_sync('isa.db')
+                self.assertEqual(cs.container_failures, 3)
+                self.assertEqual(cs.container_skips, 0)
+                self.assertEqual(1, fake_remove.call_count)
+                self.assertEqual('isa.db', fake_remove.call_args[0][0].db_file)
+
+        # Test the case where get_info returns DatabaseConnectionError
+        # error different than DB does not exist
+        with mock.patch(broker + '.get_info') as fake_get_info:
+            with mock.patch(store + '.remove_synced_container') as fake_remove:
+                fake_get_info.side_effect = DatabaseConnectionError('a', 'a')
+                cs.container_sync('isa.db')
+                self.assertEqual(cs.container_failures, 4)
+                self.assertEqual(cs.container_skips, 0)
+                self.assertEqual(0, fake_remove.call_count)
 
     def test_container_sync_not_my_db(self):
         # Db could be there due to handoff replication so test that we ignore
@@ -492,10 +533,9 @@ class TestContainerSync(unittest.TestCase):
             metadata={'x-container-sync-to': ('http://127.0.0.1/a/c', 1),
                       'x-container-sync-key': ('key', 1)},
             items_since=[{'ROWID': 1, 'name': 'o'}])
-        with nested(
-                mock.patch('swift.container.sync.ContainerBroker',
-                           lambda p: fcb),
-                mock.patch('swift.container.sync.hash_path', fake_hash_path)):
+        with mock.patch('swift.container.sync.ContainerBroker',
+                        lambda p: fcb), \
+                mock.patch('swift.container.sync.hash_path', fake_hash_path):
             cs._myips = ['10.0.0.0']    # Match
             cs._myport = 1000           # Match
             cs.allowed_sync_hosts = ['127.0.0.1']
@@ -520,10 +560,9 @@ class TestContainerSync(unittest.TestCase):
                                             'x-container-sync-key':
                                             ('key', 1)},
                                   items_since=[{'ROWID': 1, 'name': 'o'}])
-        with nested(
-                mock.patch('swift.container.sync.ContainerBroker',
-                           lambda p: fcb),
-                mock.patch('swift.container.sync.hash_path', fake_hash_path)):
+        with mock.patch('swift.container.sync.ContainerBroker',
+                        lambda p: fcb), \
+                mock.patch('swift.container.sync.hash_path', fake_hash_path):
             cs._myips = ['10.0.0.0']    # Match
             cs._myport = 1000           # Match
             cs.allowed_sync_hosts = ['127.0.0.1']
@@ -567,11 +606,10 @@ class TestContainerSync(unittest.TestCase):
                       'x-container-sync-key': ('key', 1)},
             items_since=[{'ROWID': 1, 'name': 'o', 'created_at': '1.2',
                           'deleted': True}])
-        with nested(
-                mock.patch('swift.container.sync.ContainerBroker',
-                           lambda p: fcb),
+        with mock.patch('swift.container.sync.ContainerBroker',
+                        lambda p: fcb), \
                 mock.patch('swift.container.sync.delete_object',
-                           fake_delete_object)):
+                           fake_delete_object):
             cs._myips = ['10.0.0.0']    # Match
             cs._myport = 1000           # Match
             cs.allowed_sync_hosts = ['127.0.0.1']
@@ -592,11 +630,10 @@ class TestContainerSync(unittest.TestCase):
                       'x-container-sync-key': ('key', 1)},
             items_since=[{'ROWID': 1, 'name': 'o', 'created_at': '1.2',
                           'deleted': True}])
-        with nested(
-                mock.patch('swift.container.sync.ContainerBroker',
-                           lambda p: fcb),
+        with mock.patch('swift.container.sync.ContainerBroker',
+                        lambda p: fcb), \
                 mock.patch('swift.container.sync.delete_object',
-                           lambda *x, **y: None)):
+                           lambda *x, **y: None):
             cs._myips = ['10.0.0.0']    # Match
             cs._myport = 1000           # Match
             cs.allowed_sync_hosts = ['127.0.0.1']
@@ -702,6 +739,67 @@ class TestContainerSync(unittest.TestCase):
             sync.hash_path = orig_hash_path
             sync.delete_object = orig_delete_object
 
+    def test_container_report(self):
+        container_stats = {'puts': 0,
+                           'deletes': 0,
+                           'bytes': 0}
+
+        def fake_container_sync_row(self, row, sync_to,
+                                    user_key, broker, info, realm, realm_key):
+            if 'deleted' in row:
+                container_stats['deletes'] += 1
+                return True
+
+            container_stats['puts'] += 1
+            container_stats['bytes'] += row['size']
+            return True
+
+        def fake_hash_path(account, container, obj, raw_digest=False):
+            # Ensures that no rows match for second loop, ordinal is 0 and
+            # all hashes are 1
+            return '\x01' * 16
+
+        fcb = FakeContainerBroker(
+            'path',
+            info={'account': 'a', 'container': 'c',
+                  'storage_policy_index': 0,
+                  'x_container_sync_point1': 5,
+                  'x_container_sync_point2': -1},
+            metadata={'x-container-sync-to': ('http://127.0.0.1/a/c', 1),
+                      'x-container-sync-key': ('key', 1)},
+            items_since=[{'ROWID': 1, 'name': 'o1', 'size': 0,
+                          'deleted': True},
+                         {'ROWID': 2, 'name': 'o2', 'size': 1010},
+                         {'ROWID': 3, 'name': 'o3', 'size': 0,
+                          'deleted': True},
+                         {'ROWID': 4, 'name': 'o4', 'size': 90},
+                         {'ROWID': 5, 'name': 'o5', 'size': 0}])
+
+        with mock.patch('swift.container.sync.InternalClient'), \
+                mock.patch('swift.container.sync.hash_path',
+                           fake_hash_path), \
+                mock.patch('swift.container.sync.ContainerBroker',
+                           lambda p: fcb):
+            cring = FakeRing()
+            cs = sync.ContainerSync({}, container_ring=cring,
+                                    logger=self.logger)
+            cs.container_stats = container_stats
+            cs._myips = ['10.0.0.0']    # Match
+            cs._myport = 1000           # Match
+            cs.allowed_sync_hosts = ['127.0.0.1']
+            funcType = type(sync.ContainerSync.container_sync_row)
+            cs.container_sync_row = funcType(fake_container_sync_row,
+                                             cs, sync.ContainerSync)
+            cs.container_sync('isa.db')
+            # Succeeds because no rows match
+            log_line = cs.logger.get_lines_for_level('info')[0]
+            lines = log_line.split(',')
+            self.assertTrue('sync_point2: 5', lines.pop().strip())
+            self.assertTrue('sync_point1: 5', lines.pop().strip())
+            self.assertTrue('bytes: 1100', lines.pop().strip())
+            self.assertTrue('deletes: 2', lines.pop().strip())
+            self.assertTrue('puts: 3', lines.pop().strip())
+
     def test_container_sync_row_delete(self):
         self._test_container_sync_row_delete(None, None)
 
@@ -717,6 +815,7 @@ class TestContainerSync(unittest.TestCase):
                     hex = 'abcdef'
 
             sync.uuid = FakeUUID
+            ts_data = Timestamp(1.1)
 
             def fake_delete_object(path, name=None, headers=None, proxy=None,
                                    logger=None, timeout=None):
@@ -725,12 +824,13 @@ class TestContainerSync(unittest.TestCase):
                 if realm:
                     self.assertEqual(headers, {
                         'x-container-sync-auth':
-                        'US abcdef 90e95aabb45a6cdc0892a3db5535e7f918428c90',
-                        'x-timestamp': '1.2'})
+                        'US abcdef a2401ecb1256f469494a0abcb0eb62ffa73eca63',
+                        'x-timestamp': ts_data.internal})
                 else:
                     self.assertEqual(
                         headers,
-                        {'x-container-sync-key': 'key', 'x-timestamp': '1.2'})
+                        {'x-container-sync-key': 'key',
+                         'x-timestamp': ts_data.internal})
                 self.assertEqual(proxy, 'http://proxy')
                 self.assertEqual(timeout, 5.0)
                 self.assertEqual(logger, self.logger)
@@ -741,11 +841,14 @@ class TestContainerSync(unittest.TestCase):
                 cs = sync.ContainerSync({}, container_ring=FakeRing(),
                                         logger=self.logger)
             cs.http_proxies = ['http://proxy']
-            # Success
+            # Success.
+            # simulate a row with tombstone at 1.1 and later ctype, meta times
+            created_at = ts_data.internal + '+1388+1388'  # last modified = 1.2
             self.assertTrue(cs.container_sync_row(
                 {'deleted': True,
                  'name': 'object',
-                 'created_at': '1.2'}, 'http://sync/to/path',
+                 'created_at': created_at,
+                 'size': '1000'}, 'http://sync/to/path',
                 'key', FakeContainerBroker('broker'),
                 {'account': 'a', 'container': 'c', 'storage_policy_index': 0},
                 realm, realm_key))
@@ -816,15 +919,17 @@ class TestContainerSync(unittest.TestCase):
 
     def _test_container_sync_row_put(self, realm, realm_key):
         orig_uuid = sync.uuid
-        orig_shuffle = sync.shuffle
         orig_put_object = sync.put_object
+        orig_head_object = sync.head_object
+
         try:
             class FakeUUID(object):
                 class uuid4(object):
                     hex = 'abcdef'
 
             sync.uuid = FakeUUID
-            sync.shuffle = lambda x: x
+            ts_data = Timestamp(1.1)
+            timestamp = Timestamp(1.2)
 
             def fake_put_object(sync_to, name=None, headers=None,
                                 contents=None, proxy=None, logger=None,
@@ -834,15 +939,15 @@ class TestContainerSync(unittest.TestCase):
                 if realm:
                     self.assertEqual(headers, {
                         'x-container-sync-auth':
-                        'US abcdef ef62c64bb88a33fa00722daa23d5d43253164962',
-                        'x-timestamp': '1.2',
+                        'US abcdef a5fb3cf950738e6e3b364190e246bd7dd21dad3c',
+                        'x-timestamp': timestamp.internal,
                         'etag': 'etagvalue',
                         'other-header': 'other header value',
                         'content-type': 'text/plain'})
                 else:
                     self.assertEqual(headers, {
                         'x-container-sync-key': 'key',
-                        'x-timestamp': '1.2',
+                        'x-timestamp': timestamp.internal,
                         'other-header': 'other header value',
                         'etag': 'etagvalue',
                         'content-type': 'text/plain'})
@@ -852,6 +957,8 @@ class TestContainerSync(unittest.TestCase):
                 self.assertEqual(logger, self.logger)
 
             sync.put_object = fake_put_object
+            expected_put_count = 0
+            excepted_failure_count = 0
 
             with mock.patch('swift.container.sync.InternalClient'):
                 cs = sync.ContainerSync({}, container_ring=FakeRing(),
@@ -863,20 +970,33 @@ class TestContainerSync(unittest.TestCase):
                                  '0')
                 return (200,
                         {'other-header': 'other header value',
-                         'etag': '"etagvalue"', 'x-timestamp': '1.2',
+                         'etag': '"etagvalue"',
+                         'x-timestamp': timestamp.internal,
                          'content-type': 'text/plain; swift_bytes=123'},
                         iter('contents'))
 
             cs.swift.get_object = fake_get_object
-            # Success as everything says it worked
+            # Success as everything says it worked.
+            # simulate a row with data at 1.1 and later ctype, meta times
+            created_at = ts_data.internal + '+1388+1388'  # last modified = 1.2
+
+            def fake_object_in_rcontainer(row, sync_to, user_key,
+                                          broker, realm, realm_key):
+                return False
+
+            orig_object_in_rcontainer = cs._object_in_remote_container
+            cs._object_in_remote_container = fake_object_in_rcontainer
+
             self.assertTrue(cs.container_sync_row(
                 {'deleted': False,
                  'name': 'object',
-                 'created_at': '1.2'}, 'http://sync/to/path',
+                 'created_at': created_at,
+                 'size': 50}, 'http://sync/to/path',
                 'key', FakeContainerBroker('broker'),
                 {'account': 'a', 'container': 'c', 'storage_policy_index': 0},
                 realm, realm_key))
-            self.assertEqual(cs.container_puts, 1)
+            expected_put_count += 1
+            self.assertEqual(cs.container_puts, expected_put_count)
 
             def fake_get_object(acct, con, obj, headers, acceptable_statuses):
                 self.assertEqual(headers['X-Newest'], True)
@@ -885,24 +1005,41 @@ class TestContainerSync(unittest.TestCase):
                 return (200,
                         {'date': 'date value',
                          'last-modified': 'last modified value',
-                         'x-timestamp': '1.2',
+                         'x-timestamp': timestamp.internal,
                          'other-header': 'other header value',
                          'etag': '"etagvalue"',
                          'content-type': 'text/plain; swift_bytes=123'},
                         iter('contents'))
 
             cs.swift.get_object = fake_get_object
+
             # Success as everything says it worked, also checks 'date' and
             # 'last-modified' headers are removed and that 'etag' header is
             # stripped of double quotes.
             self.assertTrue(cs.container_sync_row(
                 {'deleted': False,
                  'name': 'object',
-                 'created_at': '1.2'}, 'http://sync/to/path',
+                 'created_at': timestamp.internal,
+                 'size': 60}, 'http://sync/to/path',
                 'key', FakeContainerBroker('broker'),
                 {'account': 'a', 'container': 'c', 'storage_policy_index': 0},
                 realm, realm_key))
-            self.assertEqual(cs.container_puts, 2)
+            expected_put_count += 1
+            self.assertEqual(cs.container_puts, expected_put_count)
+
+            # Success as everything says it worked, also check that PUT
+            # timestamp equals GET timestamp when it is newer than created_at
+            # value.
+            self.assertTrue(cs.container_sync_row(
+                {'deleted': False,
+                 'name': 'object',
+                 'created_at': '1.1',
+                 'size': 60}, 'http://sync/to/path',
+                'key', FakeContainerBroker('broker'),
+                {'account': 'a', 'container': 'c', 'storage_policy_index': 0},
+                realm, realm_key))
+            expected_put_count += 1
+            self.assertEqual(cs.container_puts, expected_put_count)
 
             exc = []
 
@@ -918,11 +1055,13 @@ class TestContainerSync(unittest.TestCase):
             self.assertFalse(cs.container_sync_row(
                 {'deleted': False,
                  'name': 'object',
-                 'created_at': '1.2'}, 'http://sync/to/path',
+                 'created_at': timestamp.internal,
+                 'size': 70}, 'http://sync/to/path',
                 'key', FakeContainerBroker('broker'),
                 {'account': 'a', 'container': 'c', 'storage_policy_index': 0},
                 realm, realm_key))
-            self.assertEqual(cs.container_puts, 2)
+            self.assertEqual(cs.container_puts, expected_put_count)
+            excepted_failure_count += 1
             self.assertEqual(len(exc), 1)
             self.assertEqual(str(exc[-1]), 'test exception')
 
@@ -941,11 +1080,13 @@ class TestContainerSync(unittest.TestCase):
             self.assertFalse(cs.container_sync_row(
                 {'deleted': False,
                  'name': 'object',
-                 'created_at': '1.2'}, 'http://sync/to/path',
+                 'created_at': timestamp.internal,
+                 'size': 80}, 'http://sync/to/path',
                 'key', FakeContainerBroker('broker'),
                 {'account': 'a', 'container': 'c', 'storage_policy_index': 0},
                 realm, realm_key))
-            self.assertEqual(cs.container_puts, 2)
+            self.assertEqual(cs.container_puts, expected_put_count)
+            excepted_failure_count += 1
             self.assertEqual(len(exc), 1)
             self.assertEqual(str(exc[-1]), 'test client exception')
 
@@ -954,7 +1095,8 @@ class TestContainerSync(unittest.TestCase):
                 self.assertEqual(headers['X-Backend-Storage-Policy-Index'],
                                  '0')
                 return (200, {'other-header': 'other header value',
-                        'x-timestamp': '1.2', 'etag': '"etagvalue"'},
+                              'x-timestamp': timestamp.internal,
+                              'etag': '"etagvalue"'},
                         iter('contents'))
 
             def fake_put_object(*args, **kwargs):
@@ -966,11 +1108,14 @@ class TestContainerSync(unittest.TestCase):
             self.assertFalse(cs.container_sync_row(
                 {'deleted': False,
                  'name': 'object',
-                 'created_at': '1.2'}, 'http://sync/to/path',
+                 'created_at': timestamp.internal,
+                 'size': 90}, 'http://sync/to/path',
                 'key', FakeContainerBroker('broker'),
                 {'account': 'a', 'container': 'c', 'storage_policy_index': 0},
                 realm, realm_key))
-            self.assertEqual(cs.container_puts, 2)
+            self.assertEqual(cs.container_puts, expected_put_count)
+            excepted_failure_count += 1
+            self.assertEqual(cs.container_failures, excepted_failure_count)
             self.assertLogMessage('info', 'Unauth')
 
             def fake_put_object(*args, **kwargs):
@@ -981,11 +1126,14 @@ class TestContainerSync(unittest.TestCase):
             self.assertFalse(cs.container_sync_row(
                 {'deleted': False,
                  'name': 'object',
-                 'created_at': '1.2'}, 'http://sync/to/path',
+                 'created_at': timestamp.internal,
+                 'size': 50}, 'http://sync/to/path',
                 'key', FakeContainerBroker('broker'),
                 {'account': 'a', 'container': 'c', 'storage_policy_index': 0},
                 realm, realm_key))
-            self.assertEqual(cs.container_puts, 2)
+            self.assertEqual(cs.container_puts, expected_put_count)
+            excepted_failure_count += 1
+            self.assertEqual(cs.container_failures, excepted_failure_count)
             self.assertLogMessage('info', 'Not found', 1)
 
             def fake_put_object(*args, **kwargs):
@@ -996,16 +1144,128 @@ class TestContainerSync(unittest.TestCase):
             self.assertFalse(cs.container_sync_row(
                 {'deleted': False,
                  'name': 'object',
-                 'created_at': '1.2'}, 'http://sync/to/path',
+                 'created_at': timestamp.internal,
+                 'size': 50}, 'http://sync/to/path',
                 'key', FakeContainerBroker('broker'),
                 {'account': 'a', 'container': 'c', 'storage_policy_index': 0},
                 realm, realm_key))
-            self.assertEqual(cs.container_puts, 2)
+            self.assertEqual(cs.container_puts, expected_put_count)
+            excepted_failure_count += 1
+            self.assertEqual(cs.container_failures, excepted_failure_count)
             self.assertLogMessage('error', 'ERROR Syncing')
+
+            # Test the following cases:
+            # remote has the same date and a put doesn't take place
+            # remote has more up to date copy and a put doesn't take place
+            # head_object returns ClientException(404) and a put takes place
+            # head_object returns other ClientException put doesn't take place
+            # and we get failure
+            # head_object returns other Exception put does not take place
+            # and we get failure
+            # remote returns old copy and a put takes place
+            test_row = {'deleted': False,
+                        'name': 'object',
+                        'created_at': timestamp.internal,
+                        'etag': '1111',
+                        'size': 10}
+            test_info = {'account': 'a',
+                         'container': 'c',
+                         'storage_policy_index': 0}
+
+            actual_puts = []
+
+            def fake_put_object(*args, **kwargs):
+                actual_puts.append((args, kwargs))
+
+            def fake_head_object(*args, **kwargs):
+                return ({'x-timestamp': '1.2'}, '')
+
+            sync.put_object = fake_put_object
+            sync.head_object = fake_head_object
+            cs._object_in_remote_container = orig_object_in_rcontainer
+            self.assertTrue(cs.container_sync_row(
+                test_row, 'http://sync/to/path',
+                'key', FakeContainerBroker('broker'),
+                test_info,
+                realm, realm_key))
+            # No additional put has taken place
+            self.assertEqual(len(actual_puts), 0)
+            # No additional errors
+            self.assertEqual(cs.container_failures, excepted_failure_count)
+
+            def fake_head_object(*args, **kwargs):
+                return ({'x-timestamp': '1.3'}, '')
+
+            sync.head_object = fake_head_object
+            self.assertTrue(cs.container_sync_row(
+                test_row, 'http://sync/to/path',
+                'key', FakeContainerBroker('broker'),
+                test_info,
+                realm, realm_key))
+            # No additional put has taken place
+            self.assertEqual(len(actual_puts), 0)
+            # No additional errors
+            self.assertEqual(cs.container_failures, excepted_failure_count)
+
+            actual_puts = []
+
+            def fake_head_object(*args, **kwargs):
+                raise ClientException('test client exception', http_status=404)
+
+            sync.head_object = fake_head_object
+            self.assertTrue(cs.container_sync_row(
+                test_row, 'http://sync/to/path',
+                'key', FakeContainerBroker('broker'),
+                test_info, realm, realm_key))
+            # Additional put has taken place
+            self.assertEqual(len(actual_puts), 1)
+            # No additional errors
+            self.assertEqual(cs.container_failures, excepted_failure_count)
+
+            def fake_head_object(*args, **kwargs):
+                raise ClientException('test client exception', http_status=401)
+
+            sync.head_object = fake_head_object
+            self.assertFalse(cs.container_sync_row(
+                test_row, 'http://sync/to/path',
+                'key', FakeContainerBroker('broker'),
+                test_info, realm, realm_key))
+            # No additional put has taken place, failures increased
+            self.assertEqual(len(actual_puts), 1)
+            excepted_failure_count += 1
+            self.assertEqual(cs.container_failures, excepted_failure_count)
+
+            def fake_head_object(*args, **kwargs):
+                raise Exception()
+
+            sync.head_object = fake_head_object
+            self.assertFalse(cs.container_sync_row(
+                             test_row,
+                             'http://sync/to/path',
+                             'key', FakeContainerBroker('broker'),
+                             test_info, realm, realm_key))
+            # No additional put has taken place, failures increased
+            self.assertEqual(len(actual_puts), 1)
+            excepted_failure_count += 1
+            self.assertEqual(cs.container_failures, excepted_failure_count)
+
+            def fake_head_object(*args, **kwargs):
+                return ({'x-timestamp': '1.1'}, '')
+
+            sync.head_object = fake_head_object
+            self.assertTrue(cs.container_sync_row(
+                test_row, 'http://sync/to/path',
+                'key', FakeContainerBroker('broker'),
+                test_info, realm, realm_key))
+            # Additional put has taken place
+            self.assertEqual(len(actual_puts), 2)
+            # No additional errors
+            self.assertEqual(cs.container_failures, excepted_failure_count)
+
         finally:
             sync.uuid = orig_uuid
-            sync.shuffle = orig_shuffle
             sync.put_object = orig_put_object
+            sync.head_object = orig_head_object
 
     def test_select_http_proxy_None(self):
 
