@@ -140,19 +140,19 @@ def _create_test_rings(path, devs=None):
     ]
     intended_devs = devs or [
         {'id': 0, 'device': 'sda', 'zone': 0,
-         'region': 1, 'ip': '127.0.0.0', 'port': 6000},
+         'region': 1, 'ip': '127.0.0.0', 'port': 6200},
         {'id': 1, 'device': 'sda', 'zone': 1,
-         'region': 2, 'ip': '127.0.0.1', 'port': 6000},
+         'region': 2, 'ip': '127.0.0.1', 'port': 6200},
         {'id': 2, 'device': 'sda', 'zone': 2,
-         'region': 1, 'ip': '127.0.0.2', 'port': 6000},
+         'region': 3, 'ip': '127.0.0.2', 'port': 6200},
         {'id': 3, 'device': 'sda', 'zone': 4,
-         'region': 2, 'ip': '127.0.0.3', 'port': 6000},
+         'region': 2, 'ip': '127.0.0.3', 'port': 6200},
         {'id': 4, 'device': 'sda', 'zone': 5,
-         'region': 1, 'ip': '127.0.0.4', 'port': 6000},
+         'region': 1, 'ip': '127.0.0.4', 'port': 6200},
         {'id': 5, 'device': 'sda', 'zone': 6,
-         'region': 2, 'ip': 'fe80::202:b3ff:fe1e:8329', 'port': 6000},
+         'region': 3, 'ip': 'fe80::202:b3ff:fe1e:8329', 'port': 6200},
         {'id': 6, 'device': 'sda', 'zone': 7, 'region': 1,
-         'ip': '2001:0db8:85a3:0000:0000:8a2e:0370:7334', 'port': 6000},
+         'ip': '2001:0db8:85a3:0000:0000:8a2e:0370:7334', 'port': 6200},
     ]
     intended_part_shift = 30
     with closing(GzipFile(testgz, 'wb')) as f:
@@ -179,6 +179,10 @@ class TestObjectReplicator(unittest.TestCase):
     def setUp(self):
         utils.HASH_PATH_SUFFIX = 'endcap'
         utils.HASH_PATH_PREFIX = ''
+        # recon cache path
+        self.recon_cache = tempfile.mkdtemp()
+        rmtree(self.recon_cache, ignore_errors=1)
+        os.mkdir(self.recon_cache)
         # Setup a test ring (stolen from common/test_ring.py)
         self.testdir = tempfile.mkdtemp()
         self.devices = os.path.join(self.testdir, 'node')
@@ -191,7 +195,7 @@ class TestObjectReplicator(unittest.TestCase):
         _create_test_rings(self.testdir)
         self.logger = debug_logger('test-replicator')
         self.conf = dict(
-            bind_ip=_ips()[0], bind_port=6000,
+            bind_ip=_ips()[0], bind_port=6200,
             swift_dir=self.testdir, devices=self.devices, mount_check='false',
             timeout='300', stats_interval='1', sync_method='rsync')
         self._create_replicator()
@@ -200,6 +204,7 @@ class TestObjectReplicator(unittest.TestCase):
     def tearDown(self):
         self.assertFalse(process_errors)
         rmtree(self.testdir, ignore_errors=1)
+        rmtree(self.recon_cache, ignore_errors=1)
 
     def test_handoff_replication_setting_warnings(self):
         conf_tests = [
@@ -254,11 +259,27 @@ class TestObjectReplicator(unittest.TestCase):
         self.replicator.all_devs_info = set()
         self.df_mgr = diskfile.DiskFileManager(self.conf, self.logger)
 
+    def test_run_once_no_local_device_in_ring(self):
+        conf = dict(swift_dir=self.testdir, devices=self.devices,
+                    bind_ip='1.1.1.1', recon_cache_path=self.recon_cache,
+                    mount_check='false', timeout='300', stats_interval='1')
+        replicator = object_replicator.ObjectReplicator(conf,
+                                                        logger=self.logger)
+        replicator.run_once()
+        expected = [
+            "Can't find itself in policy with index 0 with ips 1.1.1.1 and"
+            " with port 6200 in ring file, not replicating",
+            "Can't find itself in policy with index 1 with ips 1.1.1.1 and"
+            " with port 6200 in ring file, not replicating",
+        ]
+        self.assertEqual(expected, self.logger.get_lines_for_level('error'))
+
     def test_run_once(self):
         conf = dict(swift_dir=self.testdir, devices=self.devices,
-                    bind_ip=_ips()[0],
+                    bind_ip=_ips()[0], recon_cache_path=self.recon_cache,
                     mount_check='false', timeout='300', stats_interval='1')
-        replicator = object_replicator.ObjectReplicator(conf)
+        replicator = object_replicator.ObjectReplicator(conf,
+                                                        logger=self.logger)
         was_connector = object_replicator.http_connect
         object_replicator.http_connect = mock_http_connect(200)
         cur_part = '0'
@@ -286,13 +307,16 @@ class TestObjectReplicator(unittest.TestCase):
         with _mock_process(process_arg_checker):
             replicator.run_once()
         self.assertFalse(process_errors)
+        self.assertFalse(self.logger.get_lines_for_level('error'))
         object_replicator.http_connect = was_connector
 
     # policy 1
     def test_run_once_1(self):
         conf = dict(swift_dir=self.testdir, devices=self.devices,
+                    recon_cache_path=self.recon_cache,
                     mount_check='false', timeout='300', stats_interval='1')
-        replicator = object_replicator.ObjectReplicator(conf)
+        replicator = object_replicator.ObjectReplicator(conf,
+                                                        logger=self.logger)
         was_connector = object_replicator.http_connect
         object_replicator.http_connect = mock_http_connect(200)
         cur_part = '0'
@@ -322,6 +346,7 @@ class TestObjectReplicator(unittest.TestCase):
                             side_effect=_ips):
                 replicator.run_once()
         self.assertFalse(process_errors)
+        self.assertFalse(self.logger.get_lines_for_level('error'))
         object_replicator.http_connect = was_connector
 
     def test_check_ring(self):
@@ -398,17 +423,17 @@ class TestObjectReplicator(unittest.TestCase):
             # Two disks on same IP/port
             {'id': 0, 'device': 'sda', 'zone': 0,
              'region': 1, 'ip': '1.1.1.1', 'port': 1111,
-             'replication_ip': '127.0.0.0', 'replication_port': 6000},
+             'replication_ip': '127.0.0.0', 'replication_port': 6200},
             {'id': 1, 'device': 'sdb', 'zone': 1,
              'region': 1, 'ip': '1.1.1.1', 'port': 1111,
-             'replication_ip': '127.0.0.0', 'replication_port': 6000},
+             'replication_ip': '127.0.0.0', 'replication_port': 6200},
             # Two disks on same server, different ports
             {'id': 2, 'device': 'sdc', 'zone': 2,
              'region': 2, 'ip': '1.1.1.2', 'port': 1112,
-             'replication_ip': '127.0.0.1', 'replication_port': 6000},
+             'replication_ip': '127.0.0.1', 'replication_port': 6200},
             {'id': 3, 'device': 'sdd', 'zone': 4,
              'region': 2, 'ip': '1.1.1.2', 'port': 1112,
-             'replication_ip': '127.0.0.1', 'replication_port': 6001},
+             'replication_ip': '127.0.0.1', 'replication_port': 6201},
         ]
         objects_sdb, objects_1_sdb, _, _ = self._write_disk_data('sdb')
         objects_sdc, objects_1_sdc, _, _ = self._write_disk_data('sdc')
@@ -497,22 +522,22 @@ class TestObjectReplicator(unittest.TestCase):
     def test_collect_jobs_multi_disk_diff_ports_normal(self, mock_shuffle):
         # Normally (servers_per_port=0), replication_ip AND replication_port
         # are used to determine local ring device entries.  Here we show that
-        # with bind_ip='127.0.0.1', bind_port=6000, only "sdc" is local.
+        # with bind_ip='127.0.0.1', bind_port=6200, only "sdc" is local.
         devs = [
             # Two disks on same IP/port
             {'id': 0, 'device': 'sda', 'zone': 0,
              'region': 1, 'ip': '1.1.1.1', 'port': 1111,
-             'replication_ip': '127.0.0.0', 'replication_port': 6000},
+             'replication_ip': '127.0.0.0', 'replication_port': 6200},
             {'id': 1, 'device': 'sdb', 'zone': 1,
              'region': 1, 'ip': '1.1.1.1', 'port': 1111,
-             'replication_ip': '127.0.0.0', 'replication_port': 6000},
+             'replication_ip': '127.0.0.0', 'replication_port': 6200},
             # Two disks on same server, different ports
             {'id': 2, 'device': 'sdc', 'zone': 2,
              'region': 2, 'ip': '1.1.1.2', 'port': 1112,
-             'replication_ip': '127.0.0.1', 'replication_port': 6000},
+             'replication_ip': '127.0.0.1', 'replication_port': 6200},
             {'id': 3, 'device': 'sdd', 'zone': 4,
              'region': 2, 'ip': '1.1.1.2', 'port': 1112,
-             'replication_ip': '127.0.0.1', 'replication_port': 6001},
+             'replication_ip': '127.0.0.1', 'replication_port': 6201},
         ]
         objects_sdb, objects_1_sdb, _, _ = self._write_disk_data('sdb')
         objects_sdc, objects_1_sdc, _, _ = self._write_disk_data('sdc')
@@ -576,23 +601,23 @@ class TestObjectReplicator(unittest.TestCase):
     def test_collect_jobs_multi_disk_servers_per_port(self, mock_shuffle):
         # Normally (servers_per_port=0), replication_ip AND replication_port
         # are used to determine local ring device entries.  Here we show that
-        # with servers_per_port > 0 and bind_ip='127.0.0.1', bind_port=6000,
+        # with servers_per_port > 0 and bind_ip='127.0.0.1', bind_port=6200,
         # then both "sdc" and "sdd" are local.
         devs = [
             # Two disks on same IP/port
             {'id': 0, 'device': 'sda', 'zone': 0,
              'region': 1, 'ip': '1.1.1.1', 'port': 1111,
-             'replication_ip': '127.0.0.0', 'replication_port': 6000},
+             'replication_ip': '127.0.0.0', 'replication_port': 6200},
             {'id': 1, 'device': 'sdb', 'zone': 1,
              'region': 1, 'ip': '1.1.1.1', 'port': 1111,
-             'replication_ip': '127.0.0.0', 'replication_port': 6000},
+             'replication_ip': '127.0.0.0', 'replication_port': 6200},
             # Two disks on same server, different ports
             {'id': 2, 'device': 'sdc', 'zone': 2,
              'region': 2, 'ip': '1.1.1.2', 'port': 1112,
-             'replication_ip': '127.0.0.1', 'replication_port': 6000},
+             'replication_ip': '127.0.0.1', 'replication_port': 6200},
             {'id': 3, 'device': 'sdd', 'zone': 4,
              'region': 2, 'ip': '1.1.1.2', 'port': 1112,
-             'replication_ip': '127.0.0.1', 'replication_port': 6001},
+             'replication_ip': '127.0.0.1', 'replication_port': 6201},
         ]
         objects_sdb, objects_1_sdb, _, _ = self._write_disk_data('sdb')
         objects_sdc, objects_1_sdc, _, _ = self._write_disk_data('sdc')
@@ -740,6 +765,7 @@ class TestObjectReplicator(unittest.TestCase):
         # all jobs processed!
         self.assertEqual(self.replicator.job_count,
                          self.replicator.replication_count)
+        self.assertFalse(self.replicator.handoffs_remaining)
 
         # sanity, all the handoffs suffixes we filled in were rsync'd
         found_rsync_suffix_paths = set()
@@ -767,8 +793,111 @@ class TestObjectReplicator(unittest.TestCase):
             (0, '/sda/3'): 2,
             (1, '/sda/3'): 2,
         }
-        self.assertEquals(dict(found_replicate_calls),
-                          expected_replicate_calls)
+        self.assertEqual(dict(found_replicate_calls),
+                         expected_replicate_calls)
+
+    def test_handoffs_first_mode_will_abort_if_handoffs_remaining(self):
+        # make an object in the handoff partition
+        handoff_suffix_paths = []
+        for policy in POLICIES:
+            ts = next(self.ts)
+            df = self.df_mgr.get_diskfile('sda', '1', 'a', 'c', 'o', policy)
+            with df.create() as w:
+                w.write('asdf')
+                w.put({'X-Timestamp': ts.internal})
+                w.commit(ts)
+            handoff_suffix_paths.append(os.path.dirname(df._datadir))
+        process_arg_checker = [
+            # (return_code, stdout, <each in capture rsync args>)
+            (0, '', []),
+            (1, '', []),
+            (0, '', []),
+            (0, '', []),
+            (0, '', []),
+            (0, '', []),
+        ]
+        stub_body = pickle.dumps({})
+        with _mock_process(process_arg_checker) as rsync_log, \
+            mock.patch('swift.obj.replicator.whataremyips',
+                       side_effect=_ips), \
+                mocked_http_conn(*[200] * 5, body=stub_body) as conn_log:
+            self.replicator.handoffs_first = True
+            self.replicator.replicate()
+        # stopped after handoffs!
+        self.assertEqual(1, self.replicator.handoffs_remaining)
+        self.assertEqual(8, self.replicator.job_count)
+        # in addition to the two update_deleted jobs as many as "concurrency"
+        # jobs may have been spawned into the pool before the failed
+        # update_deleted job incremented handoffs_remaining and caused the
+        # handoffs_first check to abort the current pass
+        self.assertLessEqual(self.replicator.replication_count,
+                             2 + self.replicator.concurrency)
+
+        # sanity, all the handoffs suffixes we filled in were rsync'd
+        found_rsync_suffix_paths = set()
+        expected_replicate_requests = set()
+        for subprocess_info in rsync_log:
+            local_path, remote_path = subprocess_info['rsync_args'][-2:]
+            found_rsync_suffix_paths.add(local_path)
+            if subprocess_info['ret_code'] == 0:
+                node_ip = remote_path.split(':', 1)[0]
+                expected_replicate_requests.add(node_ip)
+        self.assertEqual(set(handoff_suffix_paths), found_rsync_suffix_paths)
+        # sanity, all successful rsync nodes got REPLICATE requests
+        found_replicate_requests = set()
+        self.assertEqual(5, len(conn_log.requests))
+        for req in conn_log.requests:
+            self.assertEqual(req['method'], 'REPLICATE')
+            found_replicate_requests.add(req['ip'])
+        self.assertEqual(expected_replicate_requests,
+                         found_replicate_requests)
+
+        # and at least one partition got removed!
+        remaining_policies = []
+        for path in handoff_suffix_paths:
+            if os.path.exists(path):
+                policy = diskfile.extract_policy(path)
+                remaining_policies.append(policy)
+        self.assertEqual(len(remaining_policies), 1)
+        remaining_policy = remaining_policies[0]
+
+        # try again but with handoff_delete allowing for a single failure
+        with _mock_process(process_arg_checker) as rsync_log, \
+            mock.patch('swift.obj.replicator.whataremyips',
+                       side_effect=_ips), \
+                mocked_http_conn(*[200] * 14, body=stub_body) as conn_log:
+            self.replicator.handoff_delete = 2
+            self.replicator.replicate()
+        # all jobs processed!
+        self.assertEqual(self.replicator.job_count,
+                         self.replicator.replication_count)
+        self.assertFalse(self.replicator.handoffs_remaining)
+        # sanity, all parts got replicated
+        found_replicate_calls = defaultdict(int)
+        for req in conn_log.requests:
+            self.assertEqual(req['method'], 'REPLICATE')
+            found_replicate_key = (
+                int(req['headers']['X-Backend-Storage-Policy-Index']),
+                req['path'])
+            found_replicate_calls[found_replicate_key] += 1
+        expected_replicate_calls = {
+            (int(remaining_policy), '/sda/1/a83'): 2,
+            (0, '/sda/0'): 2,
+            (1, '/sda/0'): 2,
+            (0, '/sda/2'): 2,
+            (1, '/sda/2'): 2,
+            (0, '/sda/3'): 2,
+            (1, '/sda/3'): 2,
+        }
+        self.assertEqual(dict(found_replicate_calls),
+                         expected_replicate_calls)
+
+        # and now all handoff partitions have been rebalanced away!
+        removed_paths = set()
+        for path in handoff_suffix_paths:
+            if not os.path.exists(path):
+                removed_paths.add(path)
+        self.assertEqual(removed_paths, set(handoff_suffix_paths))
 
     def test_replicator_skips_bogus_partition_dirs(self):
         # A directory in the wrong place shouldn't crash the replicator
@@ -781,6 +910,49 @@ class TestObjectReplicator(unittest.TestCase):
         jobs = self.replicator.collect_jobs()
         self.assertEqual(len(jobs), 0)
 
+    def test_replicator_skips_rsync_temp_files(self):
+        # the empty pre-setup dirs aren't that useful to us
+        device_path = os.path.join(self.devices, 'sda')
+        rmtree(device_path, ignore_errors=1)
+        os.mkdir(device_path)
+        # create a real data file to trigger rsync
+        df = self.df_mgr.get_diskfile('sda', '0', 'a', 'c', 'o',
+                                      policy=POLICIES.legacy)
+        ts = next(self.ts)
+        with df.create() as w:
+            w.write('asdf')
+            w.put({'X-Timestamp': ts.internal})
+            w.commit(ts)
+        # pre-flight and post sync request for both other primaries
+        expected_replicate_requests = 4
+        process_arg_checker = [
+            # (return_code, stdout, <each in capture rsync args>)
+            (0, '', []),
+            (0, '', []),
+        ]
+        stub_body = pickle.dumps({})
+        with _mock_process(process_arg_checker) as rsync_log, \
+            mock.patch('swift.obj.replicator.whataremyips',
+                       side_effect=_ips), \
+                mocked_http_conn(*[200] * expected_replicate_requests,
+                                 body=stub_body) as conn_log:
+            self.replicator.replicate()
+        self.assertEqual(['REPLICATE'] * expected_replicate_requests,
+                         [r['method'] for r in conn_log.requests])
+        # expect one rsync to each other primary node
+        self.assertEqual(2, len(rsync_log))
+        expected = '--exclude=.*.[0-9a-zA-Z][0-9a-zA-Z][0-9a-zA-Z]' \
+            '[0-9a-zA-Z][0-9a-zA-Z][0-9a-zA-Z]'
+        for subprocess_info in rsync_log:
+            rsync_args = subprocess_info['rsync_args']
+            for arg in rsync_args:
+                if arg.startswith('--exclude'):
+                    self.assertEqual(arg, expected)
+                    break
+            else:
+                self.fail('Did not find --exclude argument in %r' %
+                          rsync_args)
+
     def test_replicator_removes_zbf(self):
         # After running xfs_repair, a partition directory could become a
         # zero-byte file. If this happens, the replicator should clean it
@@ -883,19 +1055,19 @@ class TestObjectReplicator(unittest.TestCase):
     def test_delete_partition_ssync_single_region(self):
         devs = [
             {'id': 0, 'device': 'sda', 'zone': 0,
-             'region': 1, 'ip': '127.0.0.0', 'port': 6000},
+             'region': 1, 'ip': '127.0.0.0', 'port': 6200},
             {'id': 1, 'device': 'sda', 'zone': 1,
-             'region': 1, 'ip': '127.0.0.1', 'port': 6000},
+             'region': 1, 'ip': '127.0.0.1', 'port': 6200},
             {'id': 2, 'device': 'sda', 'zone': 2,
-             'region': 1, 'ip': '127.0.0.2', 'port': 6000},
+             'region': 1, 'ip': '127.0.0.2', 'port': 6200},
             {'id': 3, 'device': 'sda', 'zone': 4,
-             'region': 1, 'ip': '127.0.0.3', 'port': 6000},
+             'region': 1, 'ip': '127.0.0.3', 'port': 6200},
             {'id': 4, 'device': 'sda', 'zone': 5,
-             'region': 1, 'ip': '127.0.0.4', 'port': 6000},
+             'region': 1, 'ip': '127.0.0.4', 'port': 6200},
             {'id': 5, 'device': 'sda', 'zone': 6,
-             'region': 1, 'ip': 'fe80::202:b3ff:fe1e:8329', 'port': 6000},
+             'region': 1, 'ip': 'fe80::202:b3ff:fe1e:8329', 'port': 6200},
             {'id': 6, 'device': 'sda', 'zone': 7, 'region': 1,
-             'ip': '2001:0db8:85a3:0000:0000:8a2e:0370:7334', 'port': 6000},
+             'ip': '2001:0db8:85a3:0000:0000:8a2e:0370:7334', 'port': 6200},
         ]
         _create_test_rings(self.testdir, devs=devs)
         self.conf['sync_method'] = 'ssync'
